from typing import TYPE_CHECKING

from numpy import True_, mean
if TYPE_CHECKING:
    from milcapy.model.model import SystemMilcaModel
class PlotterOptions:        # ✅✅✅
    """Clase que define las opciones de visualización"""

    def __init__(self, model: 'SystemMilcaModel'):
        """
        Inicializa las opciones de visualización con valores predeterminados.
        """
        self.model = model
        # self.values_calculator: GraphicOptionCalculator = graphic_option_calculator
        # Opciones
        # GENERALES
        self.figure_size = (10, 8)      ##
        self.dpi = 100                  ### ✅✅✅
        self.save_fig_dpi = 400         ### ✅✅✅
        self.UI_background_color = 'white' ### ✅✅✅
        self.grid = False                ##
        # 'default', 'dark_background', 'ggplot', etc.
        self.plot_style = 'default'     ##

        # Opciones para visualización de estructura
        # NODOS:
        self.UI_show_nodes = False     # ✅✅✅      #####
        self.node_size = 4              #####
        self.node_color = 'blue'            #####
        # ELEMENTOS:
        self.UI_show_members = True        #####
        self.element_line_width = 2.0   #####
        self.element_color = 'blue'    #####
        # APOYOS:
        self.UI_support = True               ###########
        self.elastic_support_label = True
        self.support_size = 0.5        #####
        self.support_color = '#00ff00'    #####
        # OPCIONES DE SELECCIÓN:
        self.highlight_selected = True
        self.selected_color = 'red'
        # ETIQUETAS DE NODOS Y ELEMENTOS:
        self.UI_node_labels = False              ###########
        self.UI_member_labels = False           ###########
        self.label_font_size = 8                ##########
        self.node_label_color = '#ed0808'        ###########
        self.member_label_color = '#26a699'     ###########


        self.UI_load = True               ###########

        # CARGAS PUNTUALES
        self.point_load = self.UI_load #True               ###########
        self.point_load_color = 'black'           ###########
        self.point_load_length_arrow = 0.2 ############
        self.point_moment_length_arrow = 0.5 ########
        # ETIQUETAS DE CARGAS PUNTUALES
        self.point_load_label = self.UI_load #True                    ##########
        self.point_load_label_color = '#ff0892'         ###########
        self.point_load_label_font_size = 8             ###########

        # CARGAS DISTRIBUIDAS
        self.distributed_load = self.UI_load #True              #######
        self.scale_dist_qload = {}                     ####
        self.scale_dist_pload = {}                     ####
        self.distributed_load_color = '#831f7a'         ########
        # ETIQUETAS DE CARGAS DISTRIBUIDAS
        self.distributed_load_label = self.UI_load #True              #######
        self.distributed_load_label_color = '#511f74'   #######
        self.distributed_load_label_font_size = 8       #######

        self.UI_reactions = False              #######
        self.reactions_color = '#b62ded'           #######

        # DEFORMADA
        self.UI_deformation_scale = {} ########### # Factor de escala para deformaciones
        self.UI_deformed = False ########### # mostrar la deformada
        self.UI_rigid_deformed = False ########### # Color para deformaciones
        self.rigid_deformed_color = '#e81f64' ########### # Color para deformaciones
        self.deformation_line_width = 2.0 ######### # Ancho de línea para deformaciones
        self.deformation_color = 'blue' # #007acc' ########### # Color para deformacione
        # CON ESTOS DATOS DE ACTUALIZA DE FORMA SIN DEFORMAR automatixcamente, y se REVIERTE CON EL BOTON DE DEFORMADA
        self.show_undeformed = True    ########## # Mostrar estructura sin deformar ################################################################################ Undeformed
        self.undeformed_color = '#cccccc'   ######### Color para estructura sin deformar
        self.disp_nodes_decimals = 7


        # ANOTACIONES DE LOS DEZPLAZAMIENTOS EN NODOS
        self.disp_nodes = True    ########## # Mostrar desplazamientos en nodos
        self.disp_nodes_color = 'black'  ########## # Color para desplazamientos en nodos
        self.disp_nodes_font_size = 8    ########## # Tamaño de fuente para desplazamientos en nodos

        # FUERZAS INTERNAS
        self.moment_on_tension_side = True     # (C| ---|Ɔ)
        self.fi_label = False                                                        ############################################################################### LABEL FI
        self.fi_line_width = 1.0           # Ancho de línea de contorno para diagramas de esfuerzos
        self.UI_axial = False
        self.axial_scale = {}
        self.UI_shear = False
        self.shear_scale = {}
        self.UI_moment = False
        self.moment_scale = {}
        self.UI_slope = False
        self.slope_scale = {}
        self.UI_deflection = False
        self.deflection_scale = {}

        colorOption = 2 # 2 = color, 1 = blanco
        self.fi_ratio_scale = 0.2
        self.positive_fill_color = '#7f7fff' if colorOption == 2 else 'white'
        self.positive_fill_alpha = 1
        self.negative_fill_color = '#ff7f7f' if colorOption == 2 else 'white'
        self.negative_fill_alpha = 1

        # RELLENOS Y CONTORNOS
        self.UI_filling_type = 'Barcolor'      # 'solid', 'barcolor'
        self.alpha_filling = 0.7               # Transparencia para relleno
        self.UI_colormap = 'jet'               # 'jet', 'viridis', 'coolwarm', etc.
        self.UI_show_colorbar = True           # Mostrar barra de colores
        self.border_contour = True             # Mostrar contornos
        self.edge_color = 'black'              # Color de bordes en contornos
        self.alpha = 0.7                       # Transparencia para contornos
        self.num_contours = 20                 # Número de niveles en contornos

        # OPCIONES DE GUARDADO
        self.save_dpi = 300                    # DPI para guardar imágenes
        self.tight_layout = True               # Ajuste automático de layout

        # OTROS:
        self.label_size = 8                    # Tamaño de fuente para etiquetas
        self.relsult_label_size = 8            # Tamaño de fuente para etiquetas de resultados



        # TOPICOS:
        self.end_length_offset = True          # Mostrar brazos en los elementos
        self.end_length_offset_color = 'black'   # Color para brazos en los elementos
        self.end_length_offset_line_width = 3.0  # Ancho de línea para brazos en los elementos

        # DESPLAMIENTOS PRESCRITOS:
        self.disp_pre_length_arrow = 1
        self.disp_pre_color = 'red'
        self.disp_pre_label_color = 'black'
        self.disp_pre_label_font_size = 8
        self.disp_pre_length_width = 2


        #! ELEMENTO CST:
        self.cst_alpha = 0.7
        self.cst_element_line_width = 0.7
        self.cst_element_label_color = 'black'
        self.cst_element_label_font_size = 8

        self.cst_edge_color = '#800000'
        self.cst_face_color = "#ff7f7f"
        self.cst_deformed_color_edge = '#800000'
        self.cst_deformed_color_face = '#ff7f7f'
        self.cst_undeformed_color_edge = '#b6f9ff'
        self.cst_undeformed_color_face = '#e3eaf3'


        #! ELEMENTO MEMBRANA Q3DOF:
        self.membrane_q3dof_alpha = 0.7
        self.membrane_q3dof_element_line_width = 0.7
        self.membrane_q3dof_element_label_color = 'black'
        self.membrane_q3dof_element_label_font_size = 8

        self.membrane_q3dof_edge_color = '#800000'
        self.membrane_q3dof_face_color = "#ff7f7f"
        self.membrane_q3dof_deformed_color_edge = '#800000'
        self.membrane_q3dof_deformed_color_face = '#ff7f7f'
        self.membrane_q3dof_undeformed_color_edge = '#b6f9ff'
        self.membrane_q3dof_undeformed_color_face = '#e3eaf3'

        #! ELEMENTO MEMBRANA Q2DOF:
        self.membrane_q2dof_alpha = 0.7
        self.membrane_q2dof_element_line_width = 0.7
        self.membrane_q2dof_element_label_color = 'black'
        self.membrane_q2dof_element_label_font_size = 8

        self.membrane_q2dof_edge_color = '#800000'
        self.membrane_q2dof_face_color = "#ff7f7f"
        self.membrane_q2dof_deformed_color_edge = '#800000'
        self.membrane_q2dof_deformed_color_face = '#ff7f7f'
        self.membrane_q2dof_undeformed_color_edge = '#b6f9ff'
        self.membrane_q2dof_undeformed_color_face = '#e3eaf3'


        #! ELEMENTO ARMADURA:
        self.truss_color = '#39a532'
        self.truss_label_color = 'black'
        self.truss_label_font_size = 8

        self.truss_deformed_color = '#39a532'
        self.truss_undeformed_color = '#b6f9ff'


        #! FRAME RELEASE:
        self.frame_release_color = '#00ff00'
        self.frame_release_point_size = 20
        self.frame_release_length_offset = 0.07 # 5% de la longitud del miembro



        self.optCargaOP = "prom" # "max", "mean", "prom"



        #! TOLERANCIAS
        self.tol = 1e-5
        self.decimals = 6

    def reset(self, pattern_name: str):
        """
        Reinicia todas las opciones a sus valores predeterminados.
        """
        self.__init__(self.model)
        self.UI_label_font_size = self.label_size
        self.point_load_label_font_size = self.label_size
        self.distributed_load_label_font_size = self.label_size
        self.disp_nodes_font_size = self.relsult_label_size
        self.reactions_font_size = self.relsult_label_size
        if self.optCargaOP == "max":
            self.load_max(pattern_name)
        elif self.optCargaOP == "mean":
            self.load_mean(pattern_name)
        elif self.optCargaOP == "prom":
            self.load_max(pattern_name, prom=True)

    def load_mean(self, pattern_name: str):
        """
        Calcula y asigna la media de los resultados para un patrón.
        """
        from numpy import nan, inf
        if self.model.membrane_q3dof != {} or self.model.membrane_q2dof != {} or self.model.csts != {}:
            fact = 0.1
        else:
            fact = 0.7
        val = self._calculate_mean(pattern_name)
        self.support_size = 0.10 * val["length_mean"] if round(val["length_mean"], 2) not in [0, None, nan, inf] else 0.4
        self.scale_dist_qload[pattern_name] = 0.07 * val["length_mean"] / val["q_mean"] if val["q_mean"] not in [0, None, nan, inf] else 0
        self.scale_dist_pload[pattern_name] = 0.02 * val["length_mean"] / val["p_mean"] if val["p_mean"] not in [0, None, nan, inf] else 0
        self.point_load_length_arrow = 0.15 * val["length_mean"]
        self.point_moment_length_arrow = 0.075 * val["length_mean"]
        self.axial_scale[pattern_name] = fact * self.fi_ratio_scale * val["length_mean"] / val["axial_mean"] if val["axial_mean"] not in [0, None, nan, inf] else 1
        self.shear_scale[pattern_name] = fact * self.fi_ratio_scale * val["length_mean"] / val["shear_mean"] if val["shear_mean"] not in [0, None, nan, inf] else 1
        self.moment_scale[pattern_name] = fact * self.fi_ratio_scale * val["length_mean"] / val["bending_mean"] if val["bending_mean"] not in [0, None, nan, inf] else 1
        self.slope_scale[pattern_name] = fact * self.fi_ratio_scale * val["length_mean"] / val["slope_mean"] if val["slope_mean"] not in [0, None, nan, inf] else 100
        # self.deflection_scale[pattern_name] = 0.15 * val["length_mean"] / val["deflection_mean"]
        self.UI_deformation_scale[pattern_name] = fact * self.fi_ratio_scale * val["length_mean"] / val["deflection_mean"] if val["deflection_mean"] not in [0, None, nan, inf] else 100

        if self.model.members == {} and self.model.membrane_q3dof == {} and self.model.membrane_q2dof == {} and self.model.csts == {}:
            axial_max = 0
            for member in self.model.trusses.values():
                axial_max = max(self.model.results[pattern_name].get_truss_axial_force(member.id)) if max(self.model.results[pattern_name].get_truss_axial_force(member.id)) > axial_max else axial_max
            def_max = max(self.model.results[pattern_name].get_model_displacements())
            self.scale_dist_pload[pattern_name] = 0.05 * val["length_max"] / axial_max if axial_max not in [0, None, nan, inf] else 1
            self.axial_scale[pattern_name] = 0.1 * val["length_max"] / axial_max if axial_max not in [0, None, nan, inf] else 0.001
            self.shear_scale[pattern_name] = 0.1
            self.moment_scale[pattern_name] = 0.1
            self.slope_scale[pattern_name] = 0.03 * val["length_max"] / def_max if def_max not in [0, None, nan, inf] else 100
            self.UI_deformation_scale[pattern_name] = 0.03 * val["length_max"] / def_max if def_max not in [0, None, nan, inf] else 100

    def load_max(self, pattern_name: str, prom: bool = False):
        """
        Calcula y asigna el máximo de los resultados para un patrón.
        """
        from numpy import nan, inf
        if self.model.membrane_q3dof != {} or self.model.membrane_q2dof != {} or self.model.csts != {}: # Si hay elementos de membrana
            fact = 0.3
        else:
            fact = 1
        val = self._calculate_max(pattern_name)
        if prom:
            lmin = val["length_min"]
            lmax = val["length_max"]
            if 0 < lmin < 0.3*lmax:
                val["length_max"] = 0.65 * (lmax + lmin) / 2
            elif 0.3*lmax < lmin < 0.4*lmax:
                val["length_max"] = 0.8 * (lmax + lmin) / 2
            elif 0.4*lmax < lmin < 0.5*lmax:
                val["length_max"] = 0.95 * (lmax + lmin) / 2
            elif 0.5*lmax < lmin < 0.6*lmax:
                val["length_max"] = 1.1 * (lmax + lmin) / 2
            elif 0.6*lmax < lmin < 0.7*lmax:
                val["length_max"] = 1.25 * (lmax + lmin) / 2
            elif 0.7*lmax < lmin < 0.8*lmax:
                val["length_max"] = 1.4 * (lmax + lmin) / 2
            elif 0.8*lmax < lmin < 0.9*lmax:
                val["length_max"] = 1.55 * (lmax + lmin) / 2
            elif 0.9*lmax < lmin < 1*lmax:
                val["length_max"] = 1.7 * (lmax + lmin) / 2
            else:
                val["length_max"] = lmax
        self.support_size = 0.10 * val["length_max"] if round(val["length_max"], 2) not in [0, None, nan, inf] else 0.4
        self.scale_dist_qload[pattern_name] = 0.15 * val["length_max"] / val["q_max"] if val["q_max"] not in [0, None, nan, inf] else 1
        self.scale_dist_pload[pattern_name] = 0.1 * val["length_max"] if val["length_max"] not in [0, None, nan, inf] else 1
        self.point_load_length_arrow = 0.15 * val["length_max"]
        self.point_moment_length_arrow = 0.075 * val["length_max"]
        self.axial_scale[pattern_name] = fact * self.fi_ratio_scale * val["length_max"] / val["axial_max"] if val["axial_max"] not in [0, None, nan, inf] else 1
        self.shear_scale[pattern_name] = fact * self.fi_ratio_scale * val["length_max"] / val["shear_max"] if val["shear_max"] not in [0, None, nan, inf] else 1
        self.moment_scale[pattern_name] = fact * self.fi_ratio_scale * val["length_max"] / val["bending_max"] if val["bending_max"] not in [0, None, nan, inf] else 1
        self.slope_scale[pattern_name] = fact * self.fi_ratio_scale * val["length_max"] / val["slope_max"] if val["slope_max"] not in [0, None, nan, inf] else 100
        # self.deflection_scale[pattern_name] = 0.15 * val["length_max"] / val["deflection_max"] if val["deflection_max"] not in [0, None, np.nan] else 0
        self.UI_deformation_scale[pattern_name] = fact * self.fi_ratio_scale * val["length_max"] / val["deflection_max"] if val["deflection_max"] not in [0, None, nan, inf] else 100

        if self.model.members == {} and self.model.membrane_q3dof == {} and self.model.membrane_q2dof == {} and self.model.csts == {}:
            axial_max = 0
            for member in self.model.trusses.values():
                axial_max = max(self.model.results[pattern_name].get_truss_axial_force(member.id)) if max(self.model.results[pattern_name].get_truss_axial_force(member.id)) > axial_max else axial_max
            def_max = max(self.model.results[pattern_name].get_model_displacements())
            self.scale_dist_pload[pattern_name] = 0.05 * val["length_max"] / axial_max if axial_max not in [0, None, nan, inf] else 1
            self.axial_scale[pattern_name] = 0.1 * val["length_max"] / axial_max if axial_max not in [0, None, nan, inf] else 0.001
            self.shear_scale[pattern_name] = 0.1
            self.moment_scale[pattern_name] = 0.1
            self.slope_scale[pattern_name] = 0.03 * val["length_max"] / def_max if def_max not in [0, None, nan, inf] else 100
            self.UI_deformation_scale[pattern_name] = 0.03 * val["length_max"] / def_max if def_max not in [0, None, nan, inf] else 100

    def _calculate_max(self, pattern_name: str):
        """
        Calcula el máximo de los resultados para un patrón.
        """
        length_max = 0
        length_max_mq3dof = 0
        length_max_mq2dof = 0
        length_max_cst = 0
        length_max_truss = 0
        length_min = 100000000000000
        q_max = 0
        p_max = 0
        deflection_max = 0
        slope_max = 0
        bending_max = 0
        shear_max = 0
        axial_max = 0
        for member, results in zip(self.model.members.values(), self.model.results[pattern_name].members.values()):
            dist_load = member.get_distributed_load(pattern_name)
            q_m = (abs(dist_load.q_i), abs(dist_load.q_j))
            p_m = (abs(dist_load.p_i), abs(dist_load.p_j))
            axial_m = abs(results["axial_forces"]).max()
            shear_m = abs(results["shear_forces"]).max()
            bending_m = abs(results["bending_moments"]).max()
            slope_m = abs(results["slopes"]).max()
            deflection_m = abs(results["deflections"]).max()

            length_max = member.length() if member.length() > length_max else length_max
            length_min = member.length() if member.length() < length_min else length_min
            q_max = max(q_m) if max(q_m) > q_max else q_max
            p_max = max(p_m) if max(p_m) > p_max else p_max
            deflection_max = deflection_m if deflection_m > deflection_max else deflection_max
            slope_max = slope_m if slope_m > slope_max else slope_max
            bending_max = bending_m if bending_m > bending_max else bending_max
            shear_max = shear_m if shear_m > shear_max else shear_max
            axial_max = axial_m if axial_m > axial_max else axial_max

        for membrane_q3dof in self.model.membrane_q3dof.values():
            minxx = min(membrane_q3dof.node1.vertex.distance_to(membrane_q3dof.node3.vertex), membrane_q3dof.node2.vertex.distance_to(membrane_q3dof.node4.vertex))
            maxxx = max(membrane_q3dof.node1.vertex.distance_to(membrane_q3dof.node3.vertex), membrane_q3dof.node2.vertex.distance_to(membrane_q3dof.node4.vertex))
            length_max_mq3dof = maxxx if maxxx > length_max_mq3dof else length_max_mq3dof
            length_min = minxx if minxx < length_min else length_min

        for membrane_q2dof in self.model.membrane_q2dof.values():
            maxxx = max(membrane_q2dof.node1.vertex.distance_to(membrane_q2dof.node3.vertex), membrane_q2dof.node2.vertex.distance_to(membrane_q2dof.node4.vertex))
            minxx = min(membrane_q2dof.node1.vertex.distance_to(membrane_q2dof.node3.vertex), membrane_q2dof.node2.vertex.distance_to(membrane_q2dof.node4.vertex))
            length_max_mq2dof = maxxx if maxxx > length_max_mq2dof else length_max_mq2dof
            length_min = minxx if minxx < length_min else length_min

        for cst in self.model.csts.values():
            maxxx = max(cst.node1.vertex.distance_to(cst.node2.vertex), cst.node2.vertex.distance_to(cst.node3.vertex), cst.node3.vertex.distance_to(cst.node1.vertex))
            minxx = min(cst.node1.vertex.distance_to(cst.node2.vertex), cst.node2.vertex.distance_to(cst.node3.vertex), cst.node3.vertex.distance_to(cst.node1.vertex))
            length_max_cst = maxxx if maxxx > length_max_cst else length_max_cst
            length_min = minxx if minxx < length_min else length_min

        for truss in self.model.trusses.values():
            length = truss.length()
            length_max_truss = length if length > length_max_truss else length_max_truss
            length_min = length if length < length_min else length_min

        return {
            "length_max": max(length_max, length_max_mq3dof, length_max_mq2dof, length_max_cst, length_max_truss),
            "length_min": length_min,
            "q_max": q_max,
            "p_max": p_max,
            "deflection_max": deflection_max,
            "slope_max": slope_max,
            "bending_max": bending_max,
            "shear_max": shear_max,
            "axial_max": axial_max
        }

    def _calculate_mean(self, pattern_name: str):
        """
        Calcula la media de los resultados para un patrón.
        """
        length_mean = 0
        length_mean_mq3dof = 0
        length_mean_mq2dof = 0
        length_mean_cst = 0
        length_mean_truss = 0
        q_mean = 0
        p_mean = 0
        deflection_mean = 0
        slope_mean = 0
        bending_mean = 0
        shear_mean = 0
        axial_mean = 0
        n = len(self.model.members) + len(self.model.membrane_q3dof) + len(self.model.membrane_q2dof) + len(self.model.csts) + len(self.model.trusses)
        for member, results in zip(self.model.members.values(), self.model.results[pattern_name].members.values()):
            dist_load = member.get_distributed_load(pattern_name)
            length_mean += member.length()
            q_mean += (abs(dist_load.q_i) + abs(dist_load.q_j))/2
            p_mean += (abs(dist_load.p_i) + abs(dist_load.p_j))/2
            deflection_mean += abs(results["deflections"]).max()
            slope_mean += abs(results["slopes"]).max()
            bending_mean += abs(results["bending_moments"]).max()
            shear_mean += abs(results["shear_forces"]).max()
            axial_mean += abs(results["axial_forces"]).max()

        for membrane_q3dof in self.model.membrane_q3dof.values():
            length_mean_mq3dof += mean([membrane_q3dof.node1.vertex.distance_to(membrane_q3dof.node3.vertex), membrane_q3dof.node2.vertex.distance_to(membrane_q3dof.node4.vertex)])

        for membrane_q2dof in self.model.membrane_q2dof.values():
            length_mean_mq2dof += mean([membrane_q2dof.node1.vertex.distance_to(membrane_q2dof.node3.vertex), membrane_q2dof.node2.vertex.distance_to(membrane_q2dof.node4.vertex)])

        for cst in self.model.csts.values():
            length_mean_cst += mean([cst.node1.vertex.distance_to(cst.node2.vertex), cst.node2.vertex.distance_to(cst.node3.vertex), cst.node3.vertex.distance_to(cst.node1.vertex)])

        for truss in self.model.trusses.values():
            length_mean_truss += truss.length()
        return {
            "length_mean": (length_mean + length_mean_mq3dof + length_mean_mq2dof + length_mean_cst + length_mean_truss)/n,
            "q_mean": q_mean/n,
            "p_mean": p_mean/n,
            "deflection_mean": deflection_mean/n,
            "slope_mean": slope_mean/n,
            "bending_mean": bending_mean/n,
            "shear_mean": shear_mean/n,
            "axial_mean": axial_mean/n
        }

    def nro_arrows(self, member_id: int):
        """
        Calcula el número de flechas para un miembro.
        """
        dx = self.support_size/0.10 * (0.10)
        if member_id in self.model.trusses:
            dxx = int(self.model.trusses[member_id].length()/dx)
        else:
            dxx = int(self.model.members[member_id].length()/dx)
        return dxx if dxx > 0 else 1



from milcapy.core.node import Node
from milcapy.section.section import ShellSection
from milcapy.utils.types import ConstitutiveModel
import numpy as np

"""
==============================================
        ELEMENTO RECTANGULAR MEMBRANA
        CON MODOS INCOMPATIBLES (Q4I)
==============================================
|| Nodos                 || 4               ||
|| DOF por nodo          || 2               ||
|| Integracion           || Reducida 4P     ||
|| Shape Function        || Bilineal        ||
|| Formulacion           || Isoparametrica  ||
|| Clase de elemento     || Serendipitos    ||
|| Estado de programacion|| Listo           ||
==============================================
"""

class MembraneQuad6I:
    """Elemento de membrana Rectangular de 4 nodos + 2 modos incompatibles adicionales para evitar el shear locking.
    NOTA ESTE ELEMENTO coverge a la solucion excata de la viga de Timoshenko (considerando deformaciones por cortante)"""
    def __init__(
        self,
        id: int,
        node1: Node,
        node2: Node,
        node3: Node,
        node4: Node,
        section: ShellSection,
        state: ConstitutiveModel,
    ) -> None:
        """
        Inicializa el elemento de membrana Rectangular de 4 nodos + 2 modos incompatibles adicionales para evitar el shear locking.

        Args:
            id (int): Identificador del elemento.
            node1 (Node): Primer nodo.
            node2 (Node): Segundo nodo.
            node3 (Node): Tercer nodo.
            node4 (Node): Cuarto nodo.
            section (ShellSection): Sección del elemento tipo area (shell).
        """
        self.id = id
        self.node1 = node1
        self.node2 = node2
        self.node3 = node3
        self.node4 = node4
        self.E = section.E()
        self.v = section.v()
        self.t = section.t
        self.section = section
        self.state = state
        self.dofs = np.concatenate((node1.dofs[:2], node2.dofs[:2], node3.dofs[:2], node4.dofs[:2]))

        r3 = np.sqrt(3)
        self.xi = [-1/r3, -1/r3, 1/r3, 1/r3]
        self.eta = [-1/r3, 1/r3, -1/r3, 1/r3]
        self.w = [1, 1, 1, 1]

    def shape_function(self, xi: float, eta: float):
        """
        Funcion de forma.
        """
        N1 = 1/4 * (1-xi) * (1-eta)
        N2 = 1/4 * (1+xi) * (1-eta)
        N3 = 1/4 * (1+xi) * (1+eta)
        N4 = 1/4 * (1-xi) * (1+eta)
        N5 = 1 + xi**2
        N6 = 1 + eta**2
        return N1, N2, N3, N4, N5, N6

    def coordinates(self, xi: float, eta: float):
        """
        Interpolacion de coordenadas del elemento.
        """
        N1,N2,N3,N4,N5,N6 = self.shape_function(xi, eta)
        x1, y1 = self.node1.vertex.coordinates
        x2, y2 = self.node2.vertex.coordinates
        x3, y3 = self.node3.vertex.coordinates
        x4, y4 = self.node4.vertex.coordinates
        x = N1*x1 + N2*x2 + N3*x3 + N4*x4
        y = N1*y1 + N2*y2 + N3*y3 + N4*y4
        return x, y

    def Ia(self) -> np.ndarray:
        """
        Matriz de unitaria modificada.
        """
        Ia = np.array([
            [1, 0, 0, 0],
            [0, 0, 0, 1],
            [0, 1, 1, 0]
        ])
        return Ia

    def J(self,xi: float, eta: float) -> np.ndarray:
        """
        Matriz de jacobiano.
        """
        x1, y1 = self.node1.vertex.coordinates
        x2, y2 = self.node2.vertex.coordinates
        x3, y3 = self.node3.vertex.coordinates
        x4, y4 = self.node4.vertex.coordinates

        J = np.array([
            [
                -x1*(1 - eta)/4 + x2*(1 - eta)/4 + x3*(eta + 1)/4 - x4*(eta + 1)/4,
                -y1*(1 - eta)/4 + y2*(1 - eta)/4 + y3*(eta + 1)/4 - y4*(eta + 1)/4
            ],
            [
                -x1*(1 - xi)/4 - x2*(xi + 1)/4 + x3*(xi + 1)/4 + x4*(1 - xi)/4,
                -y1*(1 - xi)/4 - y2*(xi + 1)/4 + y3*(xi + 1)/4 + y4*(1 - xi)/4
            ]
        ])


        return J

    def Gamma(self, xi: float, eta: float) -> np.ndarray:
        """
        Matriz de transformacion.
        """
        J = self.J(xi, eta)
        gamma = np.linalg.inv(J)
        zero = np.zeros((2,2))
        Gamma = np.vstack((
                            np.hstack((gamma, zero)),
                            np.hstack((zero, gamma))
                        ))
        return Gamma

    def dN(self, xi: float, eta: float) -> np.ndarray:
        """
        Matriz de derivadas de forma.
        """
        dN = np.array([
            [eta/4 - 1/4, 0, 1/4 - eta/4, 0, eta/4 + 1/4, 0, -eta/4 - 1/4, 0, 2*xi, 0, 0, 0],
            [xi/4 - 1/4, 0, -xi/4 - 1/4, 0, xi/4 + 1/4, 0, 1/4 - xi/4, 0, 0, 0, 2*eta, 0],
            [0, eta/4 - 1/4, 0, 1/4 - eta/4, 0, eta/4 + 1/4, 0, -eta/4 - 1/4, 0, 2*xi, 0, 0],
            [0, xi/4 - 1/4, 0, -xi/4 - 1/4, 0, xi/4 + 1/4, 0, 1/4 - xi/4, 0, 0, 0, 2*eta]
        ])

        return dN


    def B(self,xi: float, eta: float) -> np.ndarray:
        """
        Matriz de rigidez local.
        """
        dN = self.dN(xi, eta)
        Gamma = self.Gamma(xi, eta)
        Ia = self.Ia()
        B = Ia @ Gamma @ dN
        return B

    def D(self) -> np.ndarray:
        """
        Matriz constitutiva de esfuerzo plano.
        """
        v = self.section.v()
        E = self.section.E()
        # ! T E N S I O N    C O N S T A N T E
        if self.state == ConstitutiveModel.PLANE_STRESS:
            k = E/(1-v**2)
            return k*np.array([
                            [1, v, 0],
                            [v, 1, 0],
                            [0, 0, (1-v)/2],
                            ])

        # ! D E F O R M A C I O N    C O N S T A N T E
        if self.state == ConstitutiveModel.PLANE_STRAIN:
            k = (E*(1-v))/((1 + v)*(1 - 2*v))
            return k * np.array([
                [1,                 v/(1-v), 0],
                [v/(1-v),           1,       0],
                [0,                 0,       (1-2*v)/(2*(1-v))],
            ])

    def phiKi(self, xi: float, eta: float) -> np.ndarray:
        """
        Matriz funcional de rigidez.
        """
        detJ = np.linalg.det(self.J(xi, eta))
        B = self.B(xi, eta)
        D = self.D()
        t = self.t
        return B.T @ D @ B * detJ * t

    def Ki(self) -> np.ndarray:
        """
        Matriz de rigidez global.
        """
        Ki = np.zeros((12,12))
        for i in range(len(self.xi)):
            Ki += self.phiKi(self.xi[i], self.eta[i]) * self.w[i]

        # CONDENSACION ESTATICA:
        kcc = Ki[:8,:8]
        kcs = Ki[:8,8:]
        ksc = Ki[8:,:8]
        kss = Ki[8:,8:]
        try:
            Ki = kcc - kcs @ np.linalg.inv(kss) @ ksc
        except np.linalg.LinAlgError:
            Ki = kcc
        return Ki

    def get_transformation_matrix(self) -> np.ndarray:
        """
        Matriz de transformacion 12x12 de local a global.
        """
        angle = np.arctan2(self.node2.vertex.y - self.node1.vertex.y, self.node2.vertex.x - self.node1.vertex.x)

        T = np.array([
            [np.cos(angle), -np.sin(angle),       0,              0,             0,              0,              0,              0,      ],
            [np.sin(angle),  np.cos(angle),       0,              0,             0,              0,              0,              0,      ],
            [0,                   0,        np.cos(angle), -np.sin(angle),       0,              0,              0,              0,      ],
            [0,                   0,        np.sin(angle),  np.cos(angle),       0,              0,              0,              0,      ],
            [0,                   0,              0,              0,        np.cos(angle), -np.sin(angle),       0,              0,      ],
            [0,                   0,              0,              0,        np.sin(angle),  np.cos(angle),       0,              0,      ],
            [0,                   0,              0,              0,             0,              0,         np.cos(angle), -np.sin(angle)],
            [0,                   0,              0,              0,             0,              0,         np.sin(angle),  np.cos(angle)],

        ])
        return T

    def global_stiffness_matrix(self) -> np.ndarray:
        """
        Matriz de rigidez global.
        """
        # T = self.get_transformation_matrix()
        Ki = self.Ki()
        # return T @ Ki @ T.T
        return Ki


if __name__ == "__main__":
    from milcapy.utils.geometry import Vertex
    import pandas as pd
    import numpy as np
    pd.set_option('display.float_format', '{:.3f}'.format)
    E = 2.53456e6
    v = 0.2
    t = 0.25
    state = ConstitutiveModel.PLANE_STRESS
    # nd1 = Node(1, Vertex(0.0, 0.0))
    # nd2 = Node(2, Vertex(0.6, 0.0))
    # nd3 = Node(3, Vertex(0.6, 3.0))
    # nd4 = Node(4, Vertex(0.0, 3.0))
    nd1 = Node(1, Vertex(0.0, 0.0))
    nd2 = Node(2, Vertex(3.0, 0.0))
    nd3 = Node(3, Vertex(3.0, 0.6))
    nd4 = Node(4, Vertex(0.0, 0.6))
    el = MembraneQuad6I(1, nd1, nd2, nd3, nd4, E, v, t, state)
    p = 6
    # Q = np.array([-p/2, 0, -p/2, 0])
    Q = np.array([0, -p/2, 0, -p/2])
    # dofs = np.array([1, 2, 3, 4]) - 1
    dofs = np.array([3, 4, 5, 6]) - 1
    U = np.linalg.solve(el.global_stiffness_matrix()[np.ix_(dofs,dofs)], Q)
    # print(el.K())
    # print(pd.DataFrame(el.K()))
    print(pd.DataFrame(U*1000))


from typing import TYPE_CHECKING, ValuesView
from milcapy.analysis.linear_static import LinearStaticAnalysis
from milcapy.postprocess.post_processing import PostProcessing
from milcapy.core.results import Results
from milcapy.utils.types import StateType
if TYPE_CHECKING:
    from milcapy.model.model import SystemMilcaModel
    from milcapy.loads.load_pattern import LoadPattern



class AnalysisManager:
    """Clase mananger para el análisis estructural.
    maneja los tipos de análisis y opciones de análisis estructural.
    realiza el análisis estructural para todos las condiciones de carga."""

    def __init__(
        self,
        model: "SystemMilcaModel",
    ) -> None:
        """
        Inicializa el análisis estructural.

        Args:
            model: Sistema estructural a analizar.
        """
        self.model = model
        self.load_patterns = model.load_patterns
        self.results = model.results
        self.current_load_pattern: str | None = None


    def notify_all(
        self,
    ) -> None:
        """Notifica a todos los componentes del modelo."""
        # Notificar a todos los miembros del modelo
        for member in self.model.members.values():
            member.set_current_load_pattern(self.current_load_pattern)
        # Notificar a todos los cerchas del modelo
        for truss in self.model.trusses.values():
            truss.set_current_load_pattern(self.current_load_pattern)
        # Notificar a todos los nodos del modelo
        for node in self.model.nodes.values():
            node.set_current_load_pattern(self.current_load_pattern)
        # Notificar al modelo
        self.model.current_load_pattern = self.current_load_pattern

    def analyze_for_list_load_pattern(
        self,
        load_pattern_name: list[str] | None = None,
    ) -> None:
        """Ejecuta el análisis estructural para todas las condiciones de carga."""

        patterns_to_analyze: list[LoadPattern] | ValuesView[LoadPattern] = []

        if load_pattern_name is not None:
            patterns_to_analyze = [self.load_patterns[name] for name in load_pattern_name]
        else:
            patterns_to_analyze = self.load_patterns.values()

        # solucionar para cada load pattern
        for load_pattern in patterns_to_analyze:
            if load_pattern.state == StateType.INACTIVE:
                continue
            # SYSTEM NOTIFICATION: notificar a todos los componentes del modelo
            self.current_load_pattern = load_pattern.name
            self.notify_all()

            # Asignar las cargas a los nodos y miembros almacenados en el patrón de carga
            load_pattern.assign_loads_to_nodes()
            load_pattern.assign_loads_to_members()
            load_pattern.assign_prescribed_dofs_to_nodes()

            # resolver el modelo
            analysis = LinearStaticAnalysis(self.model)
            displacements, reactions = analysis.solve()

            # Crear un objeto de resultados para almacenar los resultados del análisis
            self.results[load_pattern.name] = Results()
            self.results[load_pattern.name].set_model_displacements(displacements)
            self.results[load_pattern.name].set_model_reactions(reactions)


            # crear un objeto de post-procesamiento para procesar los resultados
            post_processing = PostProcessing(self.model, self.results[load_pattern.name], self.model.postprocessing_options, load_pattern.name)
            post_processing.process_displacements_for_nodes()
            post_processing.process_reactions_for_nodes()
            post_processing.process_displacements_for_members()
            post_processing.process_internal_forces_for_members()
            post_processing.process_displacements_for_cst()
            post_processing.post_process_for_cst()
            post_processing.process_displacements_for_membrane_q3dof()
            post_processing.process_displacements_for_membrane_q2dof()
            post_processing.post_process_for_members()
            post_processing.process_displacements_for_trusses()
            post_processing.process_internal_forces_for_trusses()
            post_processing.post_process_for_trusses()

            # actualizar el estado de analisis en LoadPattern
            load_pattern.analyzed = True



from typing import TYPE_CHECKING, Tuple, Dict
import numpy as np
import time
from milcapy.loads.load import PrescribedDOF

if TYPE_CHECKING:
    from milcapy.model.model import SystemMilcaModel

class LinearStaticAnalysis:
    """
    Solucionador basado en el Método de Rigidez Directa.
    Resuelve sistemas de ecuaciones estructurales K*u = F usando métodos de solución directa.
    Usa numpy.linalg.solve(K, F)
    """

    def __init__(
        self,
        model: "SystemMilcaModel",
        ) -> None:
        """
        Inicializa el solucionador con el modelo y el método de solución.

        Args:
            model: Modelo estructural que contiene K_global y F_global.
        """
        self.model = model

        # Rendimiento y diagnóstico
        self.solution_time = 0.0
        self.assembly_time = 0.0


    def _dof_map(
        self,
    ) -> Tuple[np.ndarray, np.ndarray, Dict[int, float]]:
        """Mapea los grados de libertad del modelo para trabajar con numpy.
            [0, 1, 2, ] pertenece al nodo 1 y así sucesivamente."""
        nn = len(self.model.nodes)
        restraints = np.zeros(nn * 3, dtype=bool)

        prescribed_dofs = {} # {dof: value}

        for node in self.model.nodes.values():
            preDOF: PrescribedDOF = node.prescribed_dofs.get(self.model.current_load_pattern, PrescribedDOF())

            restraints[node.dofs[0] - 1] = node.restraints[0] #or preDOF.ux is not None
            restraints[node.dofs[1] - 1] = node.restraints[1] #or preDOF.uy is not None
            restraints[node.dofs[2] - 1] = node.restraints[2] #or preDOF.rz is not None

            if preDOF.ux is not None:
                prescribed_dofs[node.dofs[0] - 1] = preDOF.ux
            if preDOF.uy is not None:
                prescribed_dofs[node.dofs[1] - 1] = preDOF.uy
            if preDOF.rz is not None:
                prescribed_dofs[node.dofs[2] - 1] = preDOF.rz

        free_dofs = np.where(~restraints)[0]
        restrained_dofs = np.where(restraints)[0]

        return free_dofs, restrained_dofs, prescribed_dofs

    def assemble_global_load_vector(
        self,
    ) -> np.ndarray:
        """Calcula el vector de carga global del sistema.

        Returns:
            np.ndarray: Vector de carga global.
        """
        start_time = time.time()

        nn = len(self.model.nodes)
        F = np.zeros(nn * 3)

        # Asignar fuerzas nodales almacenadas en los nodos
        for node in self.model.nodes.values():
            f = node.load_vector()
            dofs = node.dofs
            F[dofs - 1] += f

        # Agregar el vector de fuerzas globales almacenadas en los miembros
        for member in self.model.members.values():
            f = member.global_load_vector()
            dofs = member.dofs
            F[dofs - 1] += f

        # Agregar cargas de los elementos CST
        for cst in self.model.csts.values():
            f = cst.get_load_vector()
            dofs = cst.dofs
            F[dofs - 1] += f

        # Agregar cargas de los elementos Truss
        for truss in self.model.trusses.values():
            f = truss.global_load_vector()
            dofs = truss.dofs
            F[dofs - 1] += f

        end_time = time.time()
        self.assembly_time += (end_time - start_time)

        return F

    def assemble_global_stiffness_matrix(
        self,
    ) -> np.ndarray:
        """Ensamblaje de la matriz de rigidez global.

        Returns:
            np.ndarray: Matriz de rigidez global.
        """
        start_time = time.time()
        nn = len(self.model.nodes)
        K = np.zeros((nn * 3, nn * 3))

        # Ensamblar la matriz de rigidez global
        # Rigides de los elmentos 1D
        for member in self.model.members.values():
            k = member.global_stiffness_matrix()
            dofs = member.dofs

            rows = np.repeat(dofs - 1, 6)
            cols = np.tile(dofs - 1, 6)
            values = k.flatten()

            for i, j, val in zip(rows, cols, values):
                K[i, j] += val

        # Rigidez de los elementos CST
        for cst in self.model.csts.values():
            k = cst.global_stiffness_matrix()
            dofs = cst.dofs

            rows = np.repeat(dofs - 1, 6)
            cols = np.tile(dofs - 1, 6)
            values = k.flatten()

            for i, j, val in zip(rows, cols, values):
                K[i, j] += val

        # Rigidez de los elementos Membrane de 3 grados de libertad por nodo
        for membQ3dof in self.model.membrane_q3dof.values():
            k = membQ3dof.global_stiffness_matrix()
            dofs = membQ3dof.dofs

            rows = np.repeat(dofs - 1, 12)
            cols = np.tile(dofs - 1, 12)
            values = k.flatten()

            for i, j, val in zip(rows, cols, values):
                K[i, j] += val


        # Rigidez de los elementos Membrana de 2 grados de libertad por nodo
        for membQ2dof in self.model.membrane_q2dof.values():
            k = membQ2dof.global_stiffness_matrix()
            dofs = membQ2dof.dofs

            rows = np.repeat(dofs - 1, 8) # 8 es el numero de grados de libertad del elemento
            cols = np.tile(dofs - 1, 8)
            values = k.flatten()

            for i, j, val in zip(rows, cols, values):
                K[i, j] += val


        # Rigidez de los elementos Armadura
        for truss in self.model.trusses.values():
            k = truss.global_stiffness_matrix()
            dofs = truss.dofs

            rows = np.repeat(dofs - 1, 4)
            cols = np.tile(dofs - 1, 4)
            values = k.flatten()

            for i, j, val in zip(rows, cols, values):
                K[i, j] += val


        # aplicar las condiciones de ejes locales de los nodos
        # for node in self.model.nodes.values():
        #     if node.local_axis is not None:
        #         dofs = node.dofs
        #         Tlg = node.local_axis.get_transformation_matrix()
        #         kfc = Tlg.T @ K[dofs - 1, dofs - 1] @ Tlg
        #         kfila = np.zeros((3, nn * 3))
        #         kcolumna = np.zeros((nn * 3, 3))
        #         for node in self.model.nodes.values():
        #             dofi = node.dofs
        #             kfila[:, dofi - 1] += Tlg.T @ K[dofs - 1, dofi - 1]
        #             kcolumna[dofi - 1, :] += K[dofi - 1, dofs - 1] @ Tlg

        #         K[dofs - 1, :] = kfila
        #         K[:, dofs - 1] = kcolumna
        #         K[dofs - 1, dofs - 1] = kfc


        K = self.apply_local_axes(K, self.model)

        # Kold = K.copy()
        # for node in self.model.nodes.values():
        #     if node.local_axis is not None:
        #         dofs = node.dofs
        #         Tlg = node.local_axis.get_transformation_matrix()

        #         # diagonal
        #         kfc = Tlg.T @ Kold[dofs - 1, dofs - 1] @ Tlg

        #         # fila y columna
        #         kfila = np.zeros((3, nn * 3))
        #         kcolumna = np.zeros((nn * 3, 3))
        #         for nodej in self.model.nodes.values():
        #             dofj = nodej.dofs
        #             kfila[:, dofj - 1] = Tlg.T @ Kold[dofs - 1, dofj - 1]
        #             kcolumna[dofj - 1, :] = Kold[dofs - 1, dofj - 1] @ Tlg

        #         # escribir en la nueva K
        #         K[dofs - 1, :] = kfila
        #         K[:, dofs - 1] = kcolumna
        #         K[dofs - 1, dofs - 1] = kfc

                # for i, j, val in zip(rows, cols, values):
                #     K[i, j] += val

        end_time = time.time()
        self.assembly_time += (end_time - start_time)

        return K

    @staticmethod
    def apply_local_axes(
        K: np.ndarray,
        model: "SystemMilcaModel"
    ) -> np.ndarray:
        """
        Aplica las transformaciones locales de cada nodo a la matriz de rigidez global K.
        """
        nn = len(model.nodes)
        ndof = nn * 3
        Kold = K.copy()
        Knew = Kold.copy()

        for node in model.nodes.values():
            if node.local_axis is None:
                continue

            dofs = node.dofs - 1  # índices base 0
            T = node.local_axis.get_transformation_matrix()

            # fila
            for nodej in model.nodes.values():
                dofj = nodej.dofs - 1
                Knew[np.ix_(dofs, dofj)] = T.T @ Kold[np.ix_(dofs, dofj)]

            # columna
            for nodej in model.nodes.values():
                dofj = nodej.dofs - 1
                Knew[np.ix_(dofj, dofs)] = Kold[np.ix_(dofj, dofs)] @ T

            # diagonal
            Knew[np.ix_(dofs, dofs)] = T.T @ Kold[np.ix_(dofs, dofs)] @ T

        return Knew



    def apply_boundary_conditions(
        self,
        K: np.ndarray,
        F: np.ndarray,
        free_dofs: np.ndarray,
        restrained_dofs: np.ndarray
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
        """Aplica las condiciones de frontera.

        Args:
            K (np.ndarray): Matriz de rigidez global.
            F (np.ndarray): Vector de fuerzas global.

        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
                - K_d: Matriz de rigidez para grados de libertad libres.
                - K_dc: Matriz de rigidez que relaciona GDL libres con restringidos.
                - K_cd: Matriz de rigidez que relaciona GDL restringidos con libres.
                - K_c: Matriz de rigidez para GDL restringidos.
                - F_d: Vector de fuerzas para GDL libres.
                - F_c: Vector de fuerzas para GDL restringidos.
        """

        # Aplicar las condiciones de apoyos elásticos
        nn = len(self.model.nodes)
        K_kk = np.zeros((nn * 3, nn * 3))
        for node in self.model.nodes.values():
            if node.elastic_supports is not None:
                kx, ky, krz = node.elastic_supports.get_elastic_supports()
                if kx is not None:
                    K_kk[node.dofs[0] - 1, node.dofs[0] - 1] += kx
                if ky is not None:
                    K_kk[node.dofs[1] - 1, node.dofs[1] - 1] += ky
                if krz is not None:
                    K_kk[node.dofs[2] - 1, node.dofs[2] - 1] += krz
        K = K + K_kk

        # Reducir la matriz de rigidez (d: desconocidos, c: conocidos)
        #     | Kd   Kdc |
        # K = |          |
        #     | Kcd   Kc |
        K_d  = K[np.ix_(free_dofs, free_dofs)]
        K_dc = K[np.ix_(free_dofs, restrained_dofs)]
        K_cd = K[np.ix_(restrained_dofs, free_dofs)]
        K_c  = K[np.ix_(restrained_dofs, restrained_dofs)]

        # Reducir el vector de fuerzas
        #     | Fd |
        # F = |    |
        #     | Fc |
        F_d = F[free_dofs]
        F_c = F[restrained_dofs]

        return K_d, K_dc, K_cd, K_c, F_d, F_c, K_kk

    def solve(
        self,
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Resuelve el sistema de ecuaciones F = KU.

        Returns:
            Tuple[np.ndarray, np.ndarray]:
                - Desplazamientos nodales
                - Reacciones en los apoyos.
        """

        start_time = time.time()
        # Obtener la matriz de rigidez global y el vector de fuerzas global
        K = self.assemble_global_stiffness_matrix()
        F = self.assemble_global_load_vector()
        self.model.global_stiffness_matrix = K
        self.model.global_load_vector = {}
        self.model.global_load_vector[self.model.current_load_pattern] = F

        # Aplicar las condiciones de frontera
        free_dofs, restrained_dofs, prescribed_dofs = self._dof_map()
        K_d, K_dc, K_cd, K_c, F_d, F_c, K_kk = self.apply_boundary_conditions(K, F, free_dofs, restrained_dofs)

        # Resolver el sistema de ecuaciones
        #     |Ud|
        # U = |  | = displacements
        #     |Uc|

        # Colocar los desplazamientos en los grados de libertad libres
        nn = len(self.model.nodes)
        # completar el vector de desplazamientos
        displacements = np.zeros(nn * 3)
        # Aplicar las condiciones de prescindidos
        displacements[list(prescribed_dofs.keys())] = list(prescribed_dofs.values())
        Uc = np.copy(displacements[restrained_dofs])
        Upc = np.copy(displacements[free_dofs])

        # resolver el sistema de ecuaciones
        if np.linalg.det(K_d) == 0:
            try:
                factorDeRegularizacion = 1e-7
                Warning(f"La matriz de rigidez global es singular, por lo que para resolver se le sumara un I[ndof. ndof] * {factorDeRegularizacion} a la matriz de rigidez global")
                K_d += np.eye(K_d.shape[0]) * factorDeRegularizacion
                U_d = np.linalg.solve(K_d, F_d - K_dc @ Uc - K_d @ Upc)
            except:
                raise ValueError("La matriz de rigidez global es singular y no se puede resolver")
        else:
            U_d = np.linalg.solve(K_d, F_d - K_dc @ Uc - K_d @ Upc)


        displacements[free_dofs] = U_d + Upc

        # Calcular las reacciones en los apoyos
        R = K_cd @ (U_d + Upc) + K_c @ Uc - F_c
        # Completar el vector de reacciones
        reactions = np.zeros(nn * 3)
        reactions[restrained_dofs] = R
        # corregir las reacciones por las condiciones de apoyos elásticos
        reactions -= K_kk @ displacements

        # Otra forma de alcular las reacciones en los apoyos, para no estar completando el vector de reacciones
        #  R  =  K_global * U_global - F_global
        # reactions = K @ displacements - F

        end_time = time.time()
        self.solution_time = (end_time - start_time)

        ## Tiempo de solucion
        print(f"Tiempo de solucion: {self.solution_time} para el Load Pattern: {self.model.current_load_pattern}")

        return displacements, reactions


from __future__ import annotations
from typing import Dict
import numpy as np

class Results:
    """Clase que almacena los resultados de un análisis de un Load Pattern.
    La estructura de los resultados es la siguiente:

    self.model: {
        "displacements": np.ndarray,
        "reactions": np.ndarray
            }

    self.nodes: {
        node_id: {
            "displacements": np.ndarray,
            "reactions": np.ndarray
                }
            }

    self.members: {
        "x_val": np.ndarray,
        member_id: {
            "displacements": np.ndarray,
            "internal_forces": np.ndarray,
            "axial_forces": np.ndarray,
            "axial_displacements": np.ndarray,
            "shear_forces": np.ndarray,
            "bending_moments": np.ndarray,
            "slopes": np.ndarray,
            "deflections": np.ndarray
                    }
            }

    self.CST: {
        "displacements": np.ndarray,
        "strains": np.ndarray,
        "stresses": np.ndarray,
        }

    self.membrane_q3dof: {
        "displacements": np.ndarray,
        "strains": np.ndarray,
        "stresses": np.ndarray,
        }

    self.membrane_q2dof: {
        "displacements": np.ndarray,
        "strains": np.ndarray,
        "stresses": np.ndarray,
        }

    self.trusses: {
        "displacements": np.ndarray,
        "internal_forces": np.ndarray,
        "axial_forces": np.ndarray,
        "axial_displacements": np.ndarray,
            }
    """

    def __init__(self):
        self.nodes: Dict[int, Dict[str, np.ndarray]] = {}
        self.members: Dict[int, Dict[str, np.ndarray]] = {}
        self.CST: Dict[int, np.ndarray] = {}
        self.membrane_q3dof: Dict[int, np.ndarray] = {}
        self.membrane_q2dof: Dict[int, np.ndarray] = {}
        self.trusses: Dict[int, Dict[str, np.ndarray]] = {}
        self.model: Dict[str, np.ndarray] = {}

    def set_model_displacements(self, displacements: np.ndarray) -> None:
        self.model["displacements"] = displacements

    def set_model_reactions(self, reactions: np.ndarray) -> None:
        self.model["reactions"] = reactions

    def set_node_displacements(self, node_id: int, displacement: np.ndarray) -> None:
        if node_id not in self.nodes:
            self.nodes[node_id] = {"displacements": np.zeros(3), "reactions": np.zeros(3)}
        self.nodes[node_id]["displacements"] = displacement

    def set_node_reactions(self, node_id: int, reaction: np.ndarray) -> None:
        if node_id not in self.nodes:
            self.nodes[node_id] = {"displacements": np.zeros(3), "reactions": np.zeros(3)}
        self.nodes[node_id]["reactions"] = reaction

    def set_member_displacements(self, member_id: int, displacement: np.ndarray) -> None:
        if member_id not in self.members:
            self.members[member_id] = {"displacements": np.zeros(6), "internal_forces": np.zeros(6)}
        self.members[member_id]["displacements"] = displacement

    def set_member_internal_forces(self, member_id: int, internal_forces: np.ndarray) -> None:
        if member_id not in self.members:
            self.members[member_id] = {"displacements": np.zeros(6), "internal_forces": np.zeros(6)}
        self.members[member_id]["internal_forces"] = internal_forces

    def set_x_val(self, member_id: int, x_val: np.ndarray) -> None:
        self.members[member_id]["x_val"] = x_val

    def set_member_axial_force(self, member_id: int, axial_force: np.ndarray) -> None:
        self.members[member_id]["axial_forces"] = axial_force

    def set_member_shear_force(self, member_id: int, shear_force: np.ndarray) -> None:
        self.members[member_id]["shear_forces"] = shear_force

    def set_member_bending_moment(self, member_id: int, bending_moment: np.ndarray) -> None:
        self.members[member_id]["bending_moments"] = bending_moment

    def set_member_deflection(self, member_id: int, deflection: np.ndarray) -> None:
        self.members[member_id]["deflections"] = deflection

    def set_member_slope(self, member_id: int, slope: np.ndarray) -> None:
        self.members[member_id]["slopes"] = slope

    def set_member_axial_displacement(self, member_id: int, axial_displacement: np.ndarray) -> None:
        self.members[member_id]["axial_displacements"] = axial_displacement

    def get_model_displacements(self) -> np.ndarray:
        return self.model["displacements"]

    def get_model_reactions(self) -> np.ndarray:
        return self.model["reactions"]

    def get_node_displacements(self, node_id: int) -> np.ndarray:
        return self.nodes[node_id]["displacements"]

    def get_node_reactions(self, node_id: int) -> np.ndarray:
        return self.nodes[node_id]["reactions"]

    def get_member_displacements(self, member_id: int) -> np.ndarray:
        return self.members[member_id]["displacements"]

    def get_member_internal_forces(self, member_id: int) -> np.ndarray:
        return self.members[member_id]["internal_forces"]

    def get_member_x_val(self, member_id: int) -> np.ndarray:
        return self.members[member_id]["x_val"]

    def get_member_axial_force(self, member_id: int) -> np.ndarray:
        return self.members[member_id]["axial_forces"]

    def get_member_shear_force(self, member_id: int) -> np.ndarray:
        return self.members[member_id]["shear_forces"]

    def get_member_bending_moment(self, member_id: int) -> np.ndarray:
        return self.members[member_id]["bending_moments"]

    def get_member_deflection(self, member_id: int) -> np.ndarray:
        return self.members[member_id]["deflections"]

    def get_member_slope(self, member_id: int) -> np.ndarray:
        return self.members[member_id]["slopes"]

    def get_member_axial_displacement(self, member_id: int) -> np.ndarray:
        return self.members[member_id]["axial_displacements"]

    def get_results_node(self, node_id: int) -> Dict[str, np.ndarray]:
        return self.nodes[node_id]

    def get_results_member(self, member_id: int) -> Dict[str, np.ndarray]:
        return self.members[member_id]

    def get_results_model(self) -> Dict[str, np.ndarray]:
        return self.model

    def get_results(self) -> Dict[str, Dict[str, np.ndarray]]:
        return self.results

    def set_cst_displacements(self, cst_id: int, displacement: np.ndarray) -> None:
        if cst_id not in self.CST:
            self.CST[cst_id] = {"displacements": np.zeros(6)}
        self.CST[cst_id]["displacements"] = displacement

    def get_cst_displacements(self, cst_id: int) -> np.ndarray:
        return self.CST[cst_id]["displacements"]

    def set_cst_strains(self, cst_id: int, strains: np.ndarray) -> None:
        self.CST[cst_id]["strains"] = strains

    def set_cst_stresses(self,  cst_id: int, stresses: np.ndarray) -> None:
            self.CST[cst_id]["stresses"] = stresses

    def get_cst_strains(self, cst_id: int) -> None:
        return self.CST[cst_id]["strains"]

    def get_cst_stresses(self, cst_id: int) -> None:
        return self.CST[cst_id]["stresses"]

    def set_membrane_q3dof_displacements(self, membrane_q3dof_id: int, displacement: np.ndarray) -> None:
        if membrane_q3dof_id not in self.membrane_q3dof:
            self.membrane_q3dof[membrane_q3dof_id] = {"displacements": np.zeros(6)}
        self.membrane_q3dof[membrane_q3dof_id]["displacements"] = displacement

    def get_membrane_q3dof_displacements(self, membrane_q3dof_id: int) -> np.ndarray:
        return self.membrane_q3dof[membrane_q3dof_id]["displacements"]

    def set_membrane_q3dof_strains(self, membrane_q3dof_id: int, strains: np.ndarray) -> None:
        pass

    def set_membrane_q3dof_stresses(self, membrane_q3dof_id: int, stresses: np.ndarray) -> None:
        pass

    def get_membrane_q3dof_strains(self, membrane_q3dof_id: int) -> None:
        pass

    def get_membrane_q3dof_stresses(self, membrane_q3dof_id: int) -> None:
        pass

    def set_membrane_q2dof_displacements(self, membrane_q2dof_id: int, displacement: np.ndarray) -> None:
        if membrane_q2dof_id not in self.membrane_q2dof:
            self.membrane_q2dof[membrane_q2dof_id] = {"displacements": np.zeros(6)}
        self.membrane_q2dof[membrane_q2dof_id]["displacements"] = displacement

    def get_membrane_q2dof_displacements(self, membrane_q2dof_id: int) -> np.ndarray:
        return self.membrane_q2dof[membrane_q2dof_id]["displacements"]

    def set_membrane_q2dof_strains(self, membrane_q2dof_id: int, strains: np.ndarray) -> None:
        pass

    def set_membrane_q2dof_stresses(self, membrane_q2dof_id: int, stresses: np.ndarray) -> None:
        pass

    def get_membrane_q2dof_strains(self, membrane_q2dof_id: int) -> None:
        pass

    def get_membrane_q2dof_stresses(self, membrane_q2dof_id: int) -> None:
        pass

    def set_truss_displacements(self, truss_id: int, displacement: np.ndarray) -> None:
        if truss_id not in self.trusses:
            self.trusses[truss_id] = {"displacements": np.zeros(6)}
        self.trusses[truss_id]["displacements"] = displacement

    def get_truss_displacements(self, truss_id: int) -> np.ndarray:
        return self.trusses[truss_id]["displacements"]

    def set_truss_axial_force(self, truss_id: int, axial_force: np.ndarray) -> None:
        self.trusses[truss_id]["axial_forces"] = axial_force

    def get_truss_axial_force(self, truss_id: int) -> np.ndarray:
        return self.trusses[truss_id]["axial_forces"]

    def set_truss_internal_forces(self, truss_id: int, internal_forces: np.ndarray) -> None:
        self.trusses[truss_id]["internal_forces"] = internal_forces

    def get_truss_internal_forces(self, truss_id: int) -> np.ndarray:
        return self.trusses[truss_id]["internal_forces"]

    def get_results_truss(self, truss_id: int) -> Dict[str, np.ndarray]:
        return self.trusses[truss_id]

    def set_truss_axial_displacement(self, truss_id: int, axial_displacement: np.ndarray) -> None:
        self.trusses[truss_id]["axial_displacements"] = axial_displacement

    def get_truss_axial_displacement(self, truss_id: int) -> np.ndarray:
        return self.trusses[truss_id]["axial_displacements"]




import matplotlib.pyplot as plt
import sys
from PySide6.QtCore import Qt
from PySide6.QtGui import QAction, QShortcut, QIcon, QDoubleValidator
from PySide6.QtWidgets import (
    QApplication, QMainWindow, QDialog, QVBoxLayout, QHBoxLayout, QGroupBox, QLabel,
    QPushButton, QCheckBox, QComboBox, QLineEdit, QColorDialog, QWidget, QWidgetAction
)
from milcapy.model.model import SystemMilcaModel
from matplotlib.backends.backend_qtagg import FigureCanvasQTAgg as FigureCanvas
import copy
import numpy as np

from milcapy.plotter.widgets import InternalForceDiagramWidget

import os
import tkinter as tk
from tkinter import filedialog


# Clase para mostrar un gráfico de Matplotlib en un widget de Qt
class MatplotlibCanvas(QWidget):
    def __init__(self, parent, model: 'SystemMilcaModel', main_window: 'MainWindow'):
        """
        Clase para mostrar un gráfico de Matplotlib en un widget de Qt.

        :param parent: El widget padre.
        :type parent: QWidget
        :param model: El modelo del sistema.
        :type model: SystemMilcaModel
        :param main_window: La ventana principal.
        :type main_window: MainWindow
        """

        super().__init__(parent)
        ############################################################
        self.model = model
        self.figure = model.plotter.figure
        self.plotter_options = model.plotter_options
        self.main_window = main_window
        self.axes = self.figure.axes  # Obtener todos los ejes
        ############################################################

        # Almacenar límites originales de cada eje
        self.original_limits = {
            ax: (ax.get_xlim(), ax.get_ylim()) for ax in self.axes}

        # Crear el lienzo de Matplotlib
        self.canvas = FigureCanvas(self.figure)
        self.canvas.draw()
        layout = QVBoxLayout()
        layout.addWidget(self.canvas)
        self.setLayout(layout)

        # Variables de paneo
        self._pan_start = None
        self.pan_active = False
        self.active_ax = None  # Eje activo

        # ==================== NODO SELECCIONABLE ====================
        # Datos de nodos (coordenadas X, Y)
        self.ax = self.model.plotter.axes
        # self.nodes = self.model.plotter.deformed_nodes[self.current_load_pattern] # {id: ax.scater(x, y)}
        self.selected_node = None  # Nodo actualmente seleccionado

        # Anotación para mostrar información del nodo seleccionado
        self.annotation = self.ax.annotate(
            "",
            xy=(0, 0),
            xytext=(10, 10),
            textcoords="offset points",
            fontsize=8,               # Tamaño de letra
            fontweight="normal",      # Negrita ("bold", "light", "normal")
            color="blue",             # Color del texto
            style="italic",           # Cursiva ("italic", "normal")
            zorder=100,
            # bbox=dict(boxstyle="round,pad=0.3", edgecolor="black", facecolor="blue"), # Estilo del fondo
            arrowprops=dict(arrowstyle="->", color="black")  # Flecha
        )

        self.annotation.set_visible(False)
        # ===============================================================

        # Conectar eventos de ratón
        self._connect_events()

    @property
    def nodes(self):
        """
        Devuelve los nodos deformados para el patrón de carga actual.

        :return: Un diccionario con los nodos deformados.
        :rtype: dict
        """
        # {id: ax.scater(x, y)}
        return self.model.plotter.deformed_nodes[self.current_load_pattern]

    @property
    def members(self):
        return self.model.plotter.members  # {id: [line2D], ...}

    @property
    def trusses(self):
        return self.model.plotter.trusses   # {id: [line2D], ...}

    @property
    def csts(self):
        return self.model.plotter.csts  # {id: [line2D, polygon2D], ...}

    @property
    def current_load_pattern(self) -> str | None:
        return self.model.current_load_pattern

    def _connect_events(self):
        self.canvas.mpl_connect('scroll_event', self._on_scroll)
        self.canvas.mpl_connect('button_press_event', self._on_button_press)
        self.canvas.mpl_connect('button_release_event', self._on_button_release)
        self.canvas.mpl_connect('motion_notify_event', self._on_mouse_move)
        self.canvas.mpl_connect('button_press_event', self._on_click)
        self.canvas.mpl_connect('button_press_event', self._on_right_click)

    def _on_click(self, event):
        """ Detecta el nodo más cercano en píxeles y activa el snap """
        if event.inaxes is None and 1:  # Clic fuera del gráfico
            return

        # Convertir coordenadas de nodos a píxeles de pantalla
        node_pixels = {node_id: self.ax.transData.transform(s.get_offsets()[0])
                       for node_id, s in self.nodes.items()}  # {id: (px_x, px_y)}

        # Coordenadas del clic en píxeles
        click_pixel = np.array([event.x, event.y])

        # Calcular distancia en píxeles entre el clic y cada nodo
        distances = {node_id: np.linalg.norm(
            pos - click_pixel) for node_id, pos in node_pixels.items()}

        # Definir umbral de snap (10 píxeles)
        snap_threshold = 10
        closest_node = min(distances, key=distances.get)  # Nodo más cercano
        closest_distance = distances[closest_node]

        if closest_distance <= snap_threshold and event.button == 1:
            # Restaurar color del nodo previamente seleccionado
            if self.selected_node is not None:
                self.nodes[self.selected_node].set_color('blue')

            # Seleccionar nuevo nodo
            self.selected_node = closest_node
            self.nodes[self.selected_node].set_color('red')  # Cambiar color

            # Actualizar anotación
            node_x, node_y = self.nodes[self.selected_node].get_offsets()[0]
            decimals = self.plotter_options.disp_nodes_decimals
            if self.model.nodes[self.selected_node].local_axis is not None:
                displacements = self.model.results[self.current_load_pattern].get_node_displacements(
                self.selected_node)
                T = self.model.nodes[self.selected_node].local_axis.get_transformation_matrix()
                displacements = T.T @ displacements
                self.annotation.set_text(
                    f"Nodo {self.selected_node}\nu1 = {displacements[0]:.{decimals}f}\nu2 = {displacements[1]:.{decimals}f}\nr3 = {displacements[2]:.{decimals}f}")
            else:
                displacements = self.model.results[self.current_load_pattern].get_node_displacements(
                self.selected_node)
                self.annotation.set_text(
                    f"Nodo {self.selected_node}\nux = {displacements[0]:.{decimals}f}\nuy = {displacements[1]:.{decimals}f}\nrz = {displacements[2]:.{decimals}f}")

            self.annotation.xy = (node_x, node_y)
            self.annotation.set_visible(True)
        elif closest_distance > snap_threshold and event.button == 1:
            # Si no hay nodos cercanos, ocultar anotación y restaurar color
            if self.selected_node is not None:
                self.nodes[self.selected_node].set_color('blue')
            self.selected_node = None
            self.annotation.set_visible(False)

        # Actualizar ploteo
        self.canvas.draw_idle()

    def _on_scroll(self, event):
        """Zoom con la rueda del ratón."""
        ax = event.inaxes
        if ax is None:
            return

        base_scale = 1.2
        scale_factor = 1 / base_scale if event.step > 0 else base_scale  # Qt usa 'step'

        x_min, x_max = ax.get_xlim()
        y_min, y_max = ax.get_ylim()

        x_center = event.xdata
        y_center = event.ydata

        x_min = x_center - (x_center - x_min) * scale_factor
        x_max = x_center + (x_max - x_center) * scale_factor
        y_min = y_center - (y_center - y_min) * scale_factor
        y_max = y_center + (y_max - y_center) * scale_factor

        ax.set_xlim(x_min, x_max)
        ax.set_ylim(y_min, y_max)
        self.canvas.draw_idle()

    def _on_button_press(self, event):
        """Activar el paneo con el botón del medio."""
        ax = event.inaxes
        if ax is not None and event.button == 2:  # Botón medio
            self._pan_start = (event.xdata, event.ydata)
            self.pan_active = True
            self.active_ax = ax
            self.setCursor(Qt.ClosedHandCursor)

    def _on_button_release(self, event):
        """Desactivar el paneo."""
        if event.button == 2:
            self.pan_active = False
            self._pan_start = None
            self.active_ax = None
            self.unsetCursor()

    def _on_mouse_move(self, event):
        """Mover el gráfico al hacer paneo."""
        if self.pan_active and self.active_ax and event.inaxes == self.active_ax and self._pan_start:
            dx = self._pan_start[0] - event.xdata
            dy = self._pan_start[1] - event.ydata

            x_min, x_max = self.active_ax.get_xlim()
            y_min, y_max = self.active_ax.get_ylim()

            self.active_ax.set_xlim(x_min + dx, x_max + dx)
            self.active_ax.set_ylim(y_min + dy, y_max + dy)
            self.canvas.draw_idle()

    def _on_right_click(self, event):
        """
        abrir ventana emergente para click derecho, una venta con las respuestas del miembro
        """
        if event.inaxes is None:  # Clic fuera del gráfico
            return
        global_contains = False

        if event.button == 3:  # Clic derecho
            for member_id, line in list(self.members.items()) + list(self.trusses.items()):
                contains, _ = line.contains(event)
                if contains:
                    global_contains = True
                    self.model.plotter.selected_member = member_id
                    # Guardar el color original y cambiar a rojo
                    if member_id in self.trusses:
                        original_color = self.model.plotter.trusses[member_id].get_color()
                        original_linewidth = self.model.plotter.trusses[member_id].get_linewidth()
                        self.model.plotter.trusses[member_id].set_color("red")
                        self.model.plotter.trusses[member_id].set_linewidth(2)
                        self.model.plotter.figure.canvas.draw_idle()
                        element = self.model.trusses[member_id]
                    else:
                        original_color = self.model.plotter.members[member_id].get_color()
                        original_linewidth = self.model.plotter.members[member_id].get_linewidth()
                        self.model.plotter.members[member_id].set_color("red")
                        self.model.plotter.members[member_id].set_linewidth(2)
                        self.model.plotter.figure.canvas.draw_idle()
                        element = self.model.members[member_id]
                    try:
                        x = self.model.results[self.current_load_pattern].get_member_x_val(member_id)
                    except:
                        x = np.linspace(0, element.length(), self.model.postprocessing_options.n)

                    # Crear ventana emergente de Tkinter y esperar a que se cierre
                    InternalForceDiagramWidget(element, self.model.plotter.diagrams, x)
                    break

            if global_contains:
                # Restaurar el color original
                plt.close("all")
                if member_id in self.trusses:
                    self.model.plotter.trusses[member_id].set_color(original_color)
                    self.model.plotter.trusses[member_id].set_linewidth(original_linewidth)
                    self.model.plotter.figure.canvas.draw_idle()
                else:
                    self.model.plotter.members[member_id].set_color(original_color)
                    self.model.plotter.members[member_id].set_linewidth(original_linewidth)
                    self.model.plotter.figure.canvas.draw_idle()


class GraphicOptionsDialog(QDialog):
    """Ventana emergente para opciones de gráfico."""

    def __init__(self, parent=None, model: 'SystemMilcaModel' = None, options=None):
        super().__init__(parent)
        """
        Inicializa la ventana de opciones de gráfico.

        Args:
            parent (QWidget, optional): El widget padre. Defaults to None.
            model (SystemMilcaModel, optional): El modelo del sistema. Defaults to None.
            options (dict, optional): Opciones de gráfico. Defaults to None.
        """
        self.setWindowTitle("Opciones de Gráfico")
        self.setMinimumSize(400, 500)
        self.model = model
        self.plotter_options = model.plotter.plotter_options
        self.options = options  # Diccionario para almacenar opciones seleccionadas

        main_layout = QVBoxLayout()

        # Secciones de opciones
        main_layout.addWidget(self.create_general_options())
        main_layout.addWidget(self.create_node_options())
        main_layout.addWidget(self.create_member_options())
        main_layout.addWidget(self.create_assignations_options())
        main_layout.addWidget(self.create_scale_options())

        # Botones Aceptar / Restaurar / Aplicar / Cancelar
        button_layout = QHBoxLayout()
        accept_button = QPushButton("Aceptar")
        restore_button = QPushButton("Restaurar Valores")
        apply_button = QPushButton("Aplicar")
        cancel_button = QPushButton("Cancelar")

        accept_button.clicked.connect(self.accept_changes)
        restore_button.clicked.connect(self.restore_defaults)
        apply_button.clicked.connect(self.apply_changes)
        cancel_button.clicked.connect(self.reject)

        button_layout.addWidget(accept_button)
        button_layout.addWidget(restore_button)
        button_layout.addWidget(apply_button)
        button_layout.addWidget(cancel_button)
        main_layout.addLayout(button_layout)

        self.setLayout(main_layout)

    @property
    def current_load_pattern(self) -> str | None:
        """
        Devuelve el Nombre del patrón de carga actual.

        Returns:
            str | None: El nombre del patrón de carga actual.
        """
        return self.model.current_load_pattern

    def create_general_options(self):
        """
        Crea las opciones generales como color de fondo.

        Returns:
            QGroupBox: El grupo de opciones generales.
        """
        group = QGroupBox("Generales")
        layout = QVBoxLayout()

        self.background_color_button = QPushButton(
            "Seleccionar Color de Fondo")
        self.background_color_button.clicked.connect(
            self._select_background_color)
        layout.addWidget(self.background_color_button)

        group.setLayout(layout)
        return group

    def create_node_options(self):
        """
        Crea las opciones para nodos.

        Returns:
            QGroupBox: El grupo de opciones para nodos.
        """
        group = QGroupBox("Nodos")
        layout = QVBoxLayout()

        self.show_nodes_checkbox = QCheckBox("Mostrar Nodos")
        self.node_labels_checkbox = QCheckBox("Mostrar Etiquetas de Nodos")

        layout.addWidget(self.show_nodes_checkbox)
        layout.addWidget(self.node_labels_checkbox)
        group.setLayout(layout)
        return group

    def create_member_options(self):
        """
        Crea las opciones para miembros (elementos).

        Returns:
            QGroupBox: El grupo de opciones para miembros.
        """
        group = QGroupBox("Miembros")
        layout = QVBoxLayout()

        self.show_members_checkbox = QCheckBox("Mostrar Miembros")
        self.member_labels_checkbox = QCheckBox(
            "Mostrar Etiquetas de Miembros")

        layout.addWidget(self.show_members_checkbox)
        layout.addWidget(self.member_labels_checkbox)
        group.setLayout(layout)
        return group

    def create_assignations_options(self):
        """
        Crea las opciones para visualización de asignaciones.

        Returns:
            QGroupBox: El grupo de opciones para visualización de asignaciones.
        """
        group = QGroupBox("Asignaciones")
        layout = QVBoxLayout()

        self.show_loads_checkbox = QCheckBox("Mostrar Cargas")
        self.show_supports_checkbox = QCheckBox("Mostrar Soportes")

        layout.addWidget(self.show_loads_checkbox)
        layout.addWidget(self.show_supports_checkbox)
        group.setLayout(layout)
        return group

    def create_scale_options(self):
        """
        Crea las opciones para visualización de la forma deformada.

        Returns:
            QGroupBox: El grupo de opciones para visualización de la forma deformada.
        """
        group = QGroupBox("Escalas")
        layout = QVBoxLayout()

        self.deformation_scale_input = QLineEdit()
        self.deformation_scale_input.setPlaceholderText(
            "Escala de deformación")
        self.deformation_scale_input.setValidator(
            QDoubleValidator(1e-9, 1e9, 20))

        layout.addWidget(QLabel("Escala de Deformación"))
        layout.addWidget(self.deformation_scale_input)

        # self.internal_forces_scale_input = QLineEdit()
        # self.internal_forces_scale_input.setPlaceholderText(
        #     "Escala de fuerzas internas")
        # self.internal_forces_scale_input.setValidator(
        #     QDoubleValidator(1e-9, 1e9, 20))

        # layout.addWidget(QLabel("Escala de Fuerzas Internas"))
        # layout.addWidget(self.internal_forces_scale_input)

        group.setLayout(layout)
        return group

    def _select_background_color(self):
        """
        Abre el selector de color para el fondo.
        """
        color = QColorDialog.getColor()
        if color.isValid():
            self.options["UI_background_color"] = color.name()

    def restore_defaults(self):
        """
        Restablece las opciones a los valores por defecto.
        """
        self.show_nodes_checkbox.setChecked(False)
        self.node_labels_checkbox.setChecked(False)
        self.show_members_checkbox.setChecked(True)
        self.member_labels_checkbox.setChecked(False)
        self.show_loads_checkbox.setChecked(True)
        self.show_supports_checkbox.setChecked(True)
        scale = self.plotter_options.UI_deformation_scale.get(self.current_load_pattern, 40)
        self.deformation_scale_input.setText(str(round(scale, 2)))
        # scale = self.plotter_options.UI_internal_forces_scale.get(self.current_load_pattern, 40)
        # self.internal_forces_scale_input.setText(str(round(scale, 2)))
        self.__recover_values()
        # ! RESTABLECE EL COLOR DE FONDO INICIAL
        self.options["UI_background_color"] = 'white'
        self.__transfer_changes()
        self.update_changer(reset=True)
        print("Restaurado a valores por defecto.")

    def accept_changes(self):
        """
        Guarda las opciones seleccionadas y las imprime.
        """
        self.__recover_values()
        self.__transfer_changes()
        self.update_changer()
        self.accept()

    def apply_changes(self):
        """
        Aplica las opciones seleccionadas pero no cierra la ventana.
        """
        self.__recover_values()
        self.__transfer_changes()
        self.update_changer()

    def __recover_values(self):
        """
        Recupera los valores de los widgets.
        """
        self.options["UI_show_nodes"] = self.show_nodes_checkbox.isChecked()
        self.options["UI_node_labels"] = self.node_labels_checkbox.isChecked()
        self.options["UI_show_members"] = self.show_members_checkbox.isChecked()
        self.options["UI_member_labels"] = self.member_labels_checkbox.isChecked()
        self.options["UI_load"] = self.show_loads_checkbox.isChecked()
        self.options["UI_support"] = self.show_supports_checkbox.isChecked()
        deformation_scale_text = self.deformation_scale_input.text()
        self.options["UI_deformation_scale"][self.current_load_pattern] = float(
            deformation_scale_text) if deformation_scale_text else 40

    def __transfer_changes(self):
        """
        Transfiere las opciones seleccionadas al objeto PlotterOptions.
        """
        op = self.plotter_options
        op.UI_background_color = self.options.get("UI_background_color", None)
        op.UI_show_nodes = self.options.get("UI_show_nodes", False)
        op.UI_show_members = self.options.get("UI_show_members", True)
        op.UI_node_labels = self.options.get("UI_node_labels", False)
        op.UI_member_labels = self.options.get("UI_member_labels", False)
        op.UI_load = self.options.get("UI_load", True)
        op.UI_support = self.options.get("UI_support", True)

    def _keep_data(self):
        """
        Asigna los valores de las opciones anteriores (mantiene) a los widgets.
        """
        self.show_nodes_checkbox.setChecked(
            self.options.get("UI_show_nodes", False))
        self.node_labels_checkbox.setChecked(
            self.options.get("UI_node_labels", False))
        self.show_members_checkbox.setChecked(
            self.options.get("UI_show_members", True))
        self.member_labels_checkbox.setChecked(
            self.options.get("UI_member_labels", False))
        self.show_loads_checkbox.setChecked(self.options.get("UI_load", True))
        self.show_supports_checkbox.setChecked(self.options.get("UI_support", True))
        self.deformation_scale_input.setText(str(self.options.get(
            "UI_deformation_scale", {}).get(self.current_load_pattern, 40)))

    def update_changer(self, reset: bool = False):
        """
        Actualiza los cambios a la figura principal y su vista inmediata.
        """
        if self.options.get("UI_background_color", None) is not None:
            self.model.plotter.change_background_color()    # cambia el color de fondo
        self.model.plotter.update_nodes()                   # actualiza los nodos
        self.model.plotter.update_members()                 # actualiza los miembros
        self.model.plotter.update_trusses()                 # actualiza los trusses
        self.model.plotter.update_cst()                     # actualiza los CSTs
        self.model.plotter.update_membrane_q3dof()          # actualiza los Membranas de 3 dof / nodo
        self.model.plotter.update_membrane_q2dof()          # actualiza los Membranas de 2 dof / nodo
        self.model.plotter.update_node_labels()             # actualiza los labels de los nodos
        self.model.plotter.update_member_labels()           # actualiza los labels de los miembros
        self.model.plotter.update_trusses_labels()          # actualiza los labels de los trusses
        self.model.plotter.update_cst_labels()              # actualiza los labels de los CSTs
        self.model.plotter.update_membrane_q3dof_labels()   # actualiza los labels de los Membranas de 3 dof / nodo
        self.model.plotter.update_membrane_q2dof_labels()   # actualiza los labels de los Membranas de 2 dof / nodo
        self.model.plotter.update_end_length_offset()       # actualiza los labels de los miembros
        self.model.plotter.update_frame_release()           # actualiza los labels de los miembros
        self.model.plotter.update_point_load()              # actualiza las cargas
        self.model.plotter.update_point_load_labels()       # actualiza los labels de las cargas
        self.model.plotter.update_distributed_loads()       # actualiza las cargas distribuidas
        self.model.plotter.update_distributed_load_labels() # actualiza los labels de las cargas distribuidas
        self.model.plotter.update_prescribed_dofs()         # actualiza los grados de libertad prescindidos
        self.model.plotter.update_prescribed_dofs_labels()  # actualiza los labels de los grados de libertad prescindidos
        self.model.plotter.update_supports()                # actualiza los soportes
        self.model.plotter.update_elastic_supports()        # actualiza los soportes elásticos
        # verificar si cambio la escala de deformación
        escala = self.options.get("UI_deformation_scale", {}).get(
            self.current_load_pattern, 40)
        if escala != self.plotter_options.UI_deformation_scale.get(self.current_load_pattern, 40):
            self.model.plotter.update_deformed(escala=escala)
            self.model.plotter.update_rigid_deformed(escala=escala)
            self.model.plotter.update_displaced_nodes(scale=escala)
            if self.model.plotter_options.UI_rigid_deformed == True or self.model.plotter_options.UI_deformed == True:
                self.model.plotter.update_supports(scale=escala)                # actualiza los soportes
                self.model.plotter.update_elastic_supports(scale=escala)        # actualiza los soportes elásticos

        if reset:
            escala = self.plotter_options.UI_deformation_scale.get(self.current_load_pattern, 40)
            self.model.plotter.update_deformed(escala=escala)
            self.model.plotter.update_rigid_deformed(escala=escala)
            self.model.plotter.update_displaced_nodes(scale=escala)
            if self.model.plotter_options.UI_rigid_deformed == True or self.model.plotter_options.UI_deformed == True:
                self.model.plotter.update_supports(scale=escala)                # actualiza los soportes
                self.model.plotter.update_elastic_supports(scale=escala)        # actualiza los soportes elásticos


# Clase para la ventana principal
class MainWindow(QMainWindow):
    def __init__(self, model: 'SystemMilcaModel'):
        """
        Inicializa la ventana principal del Visualizador del objeto "SystemMilcaModel".
        """
        super().__init__()
        self.setWindowTitle("MILCApy")
        self.setGeometry(800, 80, 800, 800)
        self.setWindowIcon(QIcon("milcapy/plotter/assets/milca.ico"))

        self.model = model

        # Crear el widget de Matplotlib
        self.plot_widget = MatplotlibCanvas(self, self.model, self)
        self.setCentralWidget(self.plot_widget)

        # Crear la barra de herramientas
        self.toolbar = self.addToolBar("Herramientas")

        # Acción para guardar IMAGENES
        action_save = QAction(QIcon("otros/logo.png"), "Guardar", self)
        action_save.triggered.connect(self.on_save_figure)
        self.toolbar.addAction(action_save)

        # SEPARADOR
        self.toolbar.addSeparator()

        # Acción para OPCIONES DE GRAFICO
        action_save = QAction(
            QIcon("otros/opciones_de_grafico.png"), "Opciones de gráfico", self)
        action_save.triggered.connect(self.abrir_opciones_grafico)
        self.toolbar.addAction(action_save)

        # SEPARADOR
        self.toolbar.addSeparator()

        # SELECCIONAR LOAD PATERNS
        self.combo = QComboBox()
        lista_patterns = list(self.model.results.keys())
        self.combo.addItems(lista_patterns)
        self.combo.currentTextChanged.connect(self.on_pattern_selected)

        # Envolver el ComboBox en un QWidgetAction
        combo_action = QWidgetAction(self)
        combo_action.setDefaultWidget(self.combo)

        # Agregar a la barra de herramientas
        self.toolbar.addAction(combo_action)
        self.toolbar.addSeparator()

        # CASILLAS DE VERIFICACION:
        self.DFA = QCheckBox("DFA")
        self.DFA.stateChanged.connect(self.mostrar_fuerzas_axiales)
        self.toolbar.addWidget(self.DFA)
        self.DFA.setChecked(self.model.plotter_options.UI_axial)
        self.toolbar.addSeparator()
        self.DFC = QCheckBox("DFC")
        self.DFC.stateChanged.connect(self.mostrar_fuerzas_cortantes)
        self.toolbar.addWidget(self.DFC)
        self.DFC.setChecked(self.model.plotter_options.UI_shear)
        self.toolbar.addSeparator()
        self.DMF = QCheckBox("DMF")
        self.DMF.stateChanged.connect(self.mostrar_momentos)
        self.toolbar.addWidget(self.DMF)
        self.DMF.setChecked(self.model.plotter_options.UI_moment)
        self.toolbar.addSeparator()
        self.REACIONES = QCheckBox("Reacciones")
        self.REACIONES.stateChanged.connect(self.mostrar_reacciones)
        self.toolbar.addWidget(self.REACIONES)
        self.REACIONES.setChecked(self.model.plotter_options.UI_reactions)
        self.toolbar.addSeparator()
        self.DEFORMADA = QCheckBox("Deformada")
        self.DEFORMADA.stateChanged.connect(self.mostrar_deformada)
        self.toolbar.addWidget(self.DEFORMADA)
        self.DEFORMADA.setChecked(self.model.plotter_options.UI_deformed)
        self.toolbar.addSeparator()
        self.DEFORMADA_RIGIDA = QCheckBox("Deformada rígida")
        self.DEFORMADA_RIGIDA.stateChanged.connect(
            self.mostrar_deformada_rigida)
        self.toolbar.addWidget(self.DEFORMADA_RIGIDA)
        self.DEFORMADA_RIGIDA.setChecked(
            self.model.plotter_options.UI_rigid_deformed)

        # Atajo de teclado (Ctrl + H) para mostrar/ocultar la barra de herramientas
        self.toggle_toolbar_shortcut = QShortcut(Qt.CTRL | Qt.Key_H, self)
        self.toggle_toolbar_shortcut.activated.connect(self.toggle_toolbar)

        # Inicialmente visible la barra de herramientas
        self.toolbar.setVisible(True)

        #####################################################
        self.options_values = {
            "UI_background_color": None,
            "UI_show_nodes": False,
            "UI_node_labels": False,
            "UI_show_members": True,
            "UI_member_labels": False,
            "UI_load": True,
            "UI_deformation_scale": copy.deepcopy(self.model.plotter_options.UI_deformation_scale),
            "UI_internal_forces_scale": 0.03,
            "UI_filling_type": "Colormap",
            "UI_colormap": "rainbow",
            "UI_support": True,
        }
        #####################################################

    def toggle_toolbar(self):
        """Activa/Desactiva la barra de herramientas"""
        self.toolbar.setVisible(not self.toolbar.isVisible())


    def on_save_figure(self):
        # Inicializar tkinter en modo oculto
        root = tk.Tk()
        root.withdraw()

        # Abrir cuadro de diálogo para elegir ruta
        path = filedialog.asksaveasfilename(
            title="Guardar figura como...",
            defaultextension=".png",
            filetypes=[("PNG", "*.png"), ("JPEG", "*.jpg"), ("Todos los archivos", "*.*")]
        )

        # Si el usuario canceló
        if not path:
            return

        # Guardar figura con matplotlib (sobrescribe si ya existe)
        self.model.plotter.figure.savefig(
            path,
            dpi=self.model.plotter_options.save_fig_dpi,
            bbox_inches="tight"
        )

        print(f"Figura guardada en: {path}")


    def abrir_opciones_grafico(self):
        """Abre la ventana emergente de opciones de gráfico"""
        dialog = GraphicOptionsDialog(self, self.model, self.options_values)
        dialog._keep_data()
        dialog.show()

    def on_pattern_selected(self, value):
        # ! NOTIFICATION
        self.model.current_load_pattern = value
        # VISUALIZACIONES AL INICIO
        self.DFA.setChecked(self.model.plotter_options.UI_axial)
        self.DFC.setChecked(self.model.plotter_options.UI_shear)
        self.DMF.setChecked(self.model.plotter_options.UI_moment)
        self.REACIONES.setChecked(self.model.plotter_options.UI_reactions)
        self.DEFORMADA.setChecked(self.model.plotter_options.UI_deformed)
        self.DEFORMADA_RIGIDA.setChecked(
            self.model.plotter_options.UI_rigid_deformed)

        # ! ACTUALIZACIONES DE VISIBILIDAD AL CAMBIAR EL PATRON
        if self.model.plotter_options.UI_deformed:
            self.mostrar_deformada(2)
        elif not self.model.plotter_options.UI_deformed:
            self.mostrar_deformada(0)
        if self.model.plotter_options.UI_rigid_deformed:
            self.mostrar_deformada_rigida(2)
        elif not self.model.plotter_options.UI_rigid_deformed:
            self.mostrar_deformada_rigida(0)
        self.model.plotter.update_change()
        if self.model.plotter_options.UI_axial:
            self.mostrar_fuerzas_axiales(2)
        elif not self.model.plotter_options.UI_axial:
            self.mostrar_fuerzas_axiales(0)
        if self.model.plotter_options.UI_shear:
            self.mostrar_fuerzas_cortantes(2)
        elif not self.model.plotter_options.UI_shear:
            self.mostrar_fuerzas_cortantes(0)
        if self.model.plotter_options.UI_moment:
            self.mostrar_momentos(2)
        elif not self.model.plotter_options.UI_moment:
            self.mostrar_momentos(0)
        if self.model.plotter_options.UI_reactions:
            self.mostrar_reacciones(2)
        elif not self.model.plotter_options.UI_reactions:
            self.mostrar_reacciones(0)
        if self.model.plotter_options.UI_support:
            self.mostrar_soportes(2)
        elif not self.model.plotter_options.UI_support:
            self.mostrar_soportes(0)

    def mostrar_fuerzas_axiales(self, state):
        """Muestra las fuerzas axiales"""
        if state == 2:
            self.model.plotter_options.UI_axial = True
            self.model.plotter.update_axial_force(visibility=True)
        elif state == 0:
            self.model.plotter_options.UI_axial = False
            self.model.plotter.update_axial_force(visibility=False)
        self.diagrams_and_deformed("F")

    def mostrar_fuerzas_cortantes(self, state):
        """Muestra las fuerzas cortantes"""
        if state == 2:
            self.model.plotter_options.UI_shear = True
            self.model.plotter.update_shear_force(visibility=True)
        elif state == 0:
            self.model.plotter_options.UI_shear = False
            self.model.plotter.update_shear_force(visibility=False)
        self.diagrams_and_deformed("F")

    def mostrar_momentos(self, state):
        """Muestra las fuerzas momentos"""
        if state == 2:
            self.model.plotter_options.UI_moment = True
            self.model.plotter.update_bending_moment(visibility=True)
        elif state == 0:
            self.model.plotter_options.UI_moment = False
            self.model.plotter.update_bending_moment(visibility=False)
        self.diagrams_and_deformed("F")

    def mostrar_reacciones(self, state):
        """Muestra las reacciones"""
        if state == 2:
            self.model.plotter_options.UI_reactions = True
            self.model.plotter.update_reactions(visibility=True)
        elif state == 0:
            self.model.plotter_options.UI_reactions = False
            self.model.plotter.update_reactions(visibility=False)

    def mostrar_deformada(self, state):
        """Muestra la deformada"""
        if state == 2: # estado activo
            scale = self.options_values.get("UI_deformation_scale", {}).get(self.model.current_load_pattern, 40)
            self.model.plotter_options.UI_deformed = True
            self.model.plotter.update_deformed()
            self.model.plotter.update_displaced_nodes()
            self.model.plotter.update_supports(scale=scale)
            self.model.plotter.update_elastic_supports(scale=scale)
        elif state == 0: # estado desactivado
            self.model.plotter_options.UI_deformed = False
            self.model.plotter.update_deformed()
            self.model.plotter.update_displaced_nodes()
            if self.model.plotter_options.UI_rigid_deformed == False and self.model.plotter_options.UI_deformed == False:
                self.model.plotter.reset_supports()
                self.model.plotter.reset_elastic_supports()
        self.diagrams_and_deformed("D")

    def mostrar_deformada_rigida(self, state):
        """Muestra la deformada rígida"""
        if state == 2: # estado activo
            scale = self.options_values.get("UI_deformation_scale", {}).get(self.model.current_load_pattern, 40)
            self.model.plotter_options.UI_rigid_deformed = True
            self.model.plotter.update_rigid_deformed()
            self.model.plotter.update_displaced_nodes()
            self.model.plotter.update_supports(scale=scale)
            self.model.plotter.update_elastic_supports(scale=scale)
        elif state == 0: # estado desactivado
            self.model.plotter_options.UI_rigid_deformed = False
            self.model.plotter.update_rigid_deformed()
            self.model.plotter.update_displaced_nodes()
            if self.model.plotter_options.UI_rigid_deformed == False and self.model.plotter_options.UI_deformed == False:
                self.model.plotter.reset_supports()
                self.model.plotter.reset_elastic_supports()
        self.diagrams_and_deformed("D")

    def mostrar_soportes(self, state):
        """Muestra los soportes"""
        if state == 2: # estado activo
            self.model.plotter_options.UI_support = True
            self.model.plotter.update_supports()
            self.model.plotter.update_elastic_supports()
        elif state == 0: # estado desactivado
            self.model.plotter_options.UI_support = False
            self.model.plotter.update_supports()
            self.model.plotter.update_elastic_supports()

    def diagrams_and_deformed(self, type):
        """Actualiza la apariencia de los miembros en función de la deformada o deformada rígida"""
        DeformOrDiagram = (self.model.plotter_options.UI_rigid_deformed == True or self.model.plotter_options.UI_deformed == True) if type == "D" else (self.model.plotter_options.UI_axial or self.model.plotter_options.UI_shear or self.model.plotter_options.UI_moment)
        if DeformOrDiagram:
            self.model.plotter_options.UI_load = False
            if type == "D": # mientras se muestra la deformada
                if self.model.plotter_options.show_undeformed: # si se muestra la NO DEFORMADA
                    self.model.plotter.update_members(color=self.model.plotter_options.undeformed_color) # COLOR DE LA NO DEFORMADA CUANDO LA DEFORMADA SE MUESTRA
                    self.model.plotter.update_trusses(color=self.model.plotter_options.undeformed_color)
                    self.model.plotter.update_cst(color_edge=self.model.plotter_options.cst_undeformed_color_edge, color_face=self.model.plotter_options.cst_undeformed_color_face)
                    self.model.plotter.update_membrane_q3dof(color_edge=self.model.plotter_options.membrane_q3dof_undeformed_color_edge, color_face=self.model.plotter_options.membrane_q3dof_undeformed_color_face)
                    self.model.plotter.update_membrane_q2dof(color_edge=self.model.plotter_options.membrane_q2dof_undeformed_color_edge, color_face=self.model.plotter_options.membrane_q2dof_undeformed_color_face)
                    self.model.plotter.update_end_length_offset(color=self.model.plotter_options.undeformed_color)
                    self.model.plotter.update_frame_release(color=self.model.plotter_options.undeformed_color)
                else: # si no se muestra la NO DEFORMADA
                    self.model.plotter_options.UI_show_members = False
                    self.options_values["UI_show_members"] = False
                    self.model.plotter.update_members()
                    self.model.plotter.update_trusses()
                    self.model.plotter.update_cst(color_edge=self.model.plotter_options.cst_edge_color, color_face=self.model.plotter_options.cst_face_color)
                    self.model.plotter.update_membrane_q3dof(color_edge=self.model.plotter_options.membrane_q3dof_edge_color, color_face=self.model.plotter_options.membrane_q3dof_face_color)
                    self.model.plotter.update_membrane_q2dof(color_edge=self.model.plotter_options.membrane_q2dof_edge_color, color_face=self.model.plotter_options.membrane_q2dof_face_color)
                    self.model.plotter.update_end_length_offset()
                    self.model.plotter.update_frame_release()
                    self.model.plotter_options.UI_show_members = True
                    self.options_values["UI_show_members"] = True

        else:
            self.model.plotter_options.UI_load = True
            if type == "D":
                self.model.plotter.update_members(color=self.model.plotter_options.element_color)
                self.model.plotter.update_trusses(color=self.model.plotter_options.truss_color)
                self.model.plotter.update_cst(color_edge=self.model.plotter_options.cst_edge_color, color_face=self.model.plotter_options.cst_face_color)
                self.model.plotter.update_membrane_q3dof(color_edge=self.model.plotter_options.membrane_q3dof_edge_color, color_face=self.model.plotter_options.membrane_q3dof_face_color)
                self.model.plotter.update_membrane_q2dof(color_edge=self.model.plotter_options.membrane_q2dof_edge_color, color_face=self.model.plotter_options.membrane_q2dof_face_color)
                self.model.plotter.update_end_length_offset(color=self.model.plotter_options.end_length_offset_color)
                self.model.plotter.update_frame_release(color=self.model.plotter_options.frame_release_color)
                self.options_values["UI_show_members"] = True

        if self.options_values["UI_load"] == True:
            self.model.plotter.update_distributed_loads()
            self.model.plotter.update_point_load()
            self.model.plotter.update_distributed_load_labels()
            self.model.plotter.update_point_load_labels()
            self.model.plotter.update_prescribed_dofs()
            self.model.plotter.update_prescribed_dofs_labels()



        #! AJUSTES FORZADOS:
        self.model.plotter.update_cst_labels()
        self.model.plotter.update_membrane_q3dof_labels()
        self.model.plotter.update_membrane_q2dof_labels()
        self.model.plotter.update_trusses_labels()



def main_window(model: 'SystemMilcaModel'):
    """
    Inicializa y ejecuta la ventana principal del Visualizador del objeto "SystemMilcaModel".
    """
    app = QApplication.instance()
    if app is None:  # Si no existe una instancia, se crea
        app = QApplication(sys.argv)
    window = MainWindow(model)
    window.show()
    sys.exit(app.exec())



import numpy as np
from functools import lru_cache

CACHE_COPIES = None

# ==========================================================================================================
# ===================================   MATRICES DE RIGIDEZ   ==============================================
# ==========================================================================================================



@lru_cache(maxsize=CACHE_COPIES)
def local_stiffness_matrix(
    E: float,  # Módulo de Young
    I: float,  # Momento de inercia
    A: float,  # Área de la sección
    L: float,  # Longitud del elemento
    le: float,  # Longitud de la parte flexible
    phi: float,  # Aporte de cortante
) -> np.ndarray:
    """Matriz de rigidez local de una viga.

    Args:
        E (float): Módulo de elasticidad.
        I (float): Momento de inercia.
        A (float): Área transversal.
        L (float): Longitud del elemento.
        phi (float): Aporte de cortante.

    Returns:
        np.ndarray: Matriz de rigidez 6x6.
    """
    k11 = E * A / L
    k22 = 12 * E * I / (le**3 * (1 + phi))
    k23 = 6 * E * I / (le**2 * (1 + phi))
    k33 = (4 + phi) * E * I / (le * (1 + phi))
    k66 = (2 - phi) * E * I / (le * (1 + phi))

    return np.array([
        [ k11,   0,    0,   -k11,   0,    0   ],
        [  0,   k22,  k23,    0,   -k22,  k23 ],
        [  0,   k23,  k33,    0,   -k23,  k66 ],
        [-k11,   0,    0,    k11,   0,    0   ],
        [  0,  -k22, -k23,    0,    k22, -k23 ],
        [  0,   k23,  k66,    0,   -k23,  k33 ]
    ], dtype=np.float64)

# ==========================================================================================================
# ===================================   MATRICES DE TRANFORMACION   ========================================
# ==========================================================================================================


@lru_cache(maxsize=CACHE_COPIES)
def transformation_matrix(
    angle: float
) -> np.ndarray:
    """Matriz de transformación para un ángulo dado del sistema local al sistema global.

    Args:
        angle (float): Ángulo de rotación en radianes.

    Returns:
        np.ndarray: Matriz de transformación.
    """
    s, c = np.sin(angle), np.cos(angle)
    return np.array([
        [ c,  s, 0,  0, 0, 0 ],
        [-s,  c, 0,  0, 0, 0 ],
        [ 0,  0, 1,  0, 0, 0 ],
        [ 0,  0, 0,  c, s, 0 ],
        [ 0,  0, 0, -s, c, 0 ],
        [ 0,  0, 0,  0, 0, 1 ]
    ])

@lru_cache(maxsize=CACHE_COPIES)
def length_offset_transformation_matrix(la: float, lb: float) -> np.ndarray:
    """
    Calcula la matriz de tranformacion para miembros con brazos rigidos.

    La matriz de tranformacion se utiliza para transformar las cargas
    distribuidas en los brazos a los extremos de la barra.

    Parameters
    ----------
    la : float
        Longitud del brazo izquierdo.
    lb : float
        Longitud del brazo derecho.

    Returns
    -------
    np.ndarray
        Matriz de tranformacion 6x6.
    """
    return np.array([
        [1, 0, 0,  0, 0,   0],
        [0, 1, la, 0, 0,   0],
        [0, 0, 1,  0, 0,   0],
        [0, 0, 0,  1, 0,   0],
        [0, 0, 0,  0, 1, -lb],
        [0, 0, 0,  0, 0,   1]
    ], dtype=np.float64)




# ==========================================================================================================
# ===================================   VECTORES DE CARGAS   ===============================================
# ==========================================================================================================

@lru_cache(maxsize=CACHE_COPIES)
def q_phi(L, phi, qi, qj, pi, pj):
    """
    Calcula el vector de carga para un miembro de longitud L con deformaciones de corte
    y cargas distribuidas qi y qj.

    Args:
        qi (float): Intensidad de carga distribuida transversal en el nodo inicial.
        qj (float): Intensidad de carga distribuida transversal en el nodo final.
        pi (float): Intensidad de carga distribuida axial en el nodo inicial.
        pj (float): Intensidad de carga distribuida axial en el nodo final.
        L (float): Longitud del miembro.
        phi (float): Aporte de cortante.

    Returns:
        np.ndarray: Vector de carga.
    """
    if L == 0:
        return np.zeros(6)
    a = (qj - qi) / L
    b = qi

    N1 = (2 * pi + pj) * L / 6
    V1 = b * L / 2 + a * L**2 / 60 * ((10 * phi + 9) / (phi + 1))
    M1 = b * L**2 / 12 + a * L**3 / 120 * ((5 * phi + 4) / (phi + 1))
    N2 = (pi + 2 * pj) * L / 6
    V2 = b * L / 2 + a * L**2 / 60 * ((20 * phi + 21) / (phi + 1))
    M2 = -(b * L**2 / 12 + a * L**3 / 120 * ((5 * phi + 6) / (phi + 1)))
    return np.array([N1, V1, M1, N2, V2, M2])

@lru_cache(maxsize=CACHE_COPIES)
def length_offset_q(L, phi, qi, qj, pi, pj, la, lb, qla, qlb):
    """
    Calcula el vector de cargas considerando brazos rigidos.

    Primero se calculan las cargas distribuidas en los extremos de los brazos
    rigidos y luego se suman las cargas distribuidas en la parte flexible y
    en los extremos de los brazos rigidos.

    Args:
        L (float): Longitud del miembro.
        phi (float): Aporte de cortante.
        qi (float): Intensidad de carga distribuida transversal en el nodo inicial.
        qj (float): Intensidad de carga distribuida transversal en el nodo final.
        la (float): Longitud del brazo izquierdo.
        lb (float): Longitud del brazo derecho.

    Returns:
        np.ndarray: Vector de carga.
    """

    a = (qj - qi) / L
    b = qi

    c = (pj - pi) / L
    d = pi

    qa = a*la + b
    qb = a*(L -lb) + b

    pa = c*la + d
    pb = c*(L -lb) + d

    ####### CORRECIION ADICIONAL: el usuario al darle qla, qlb == False
    # indica que los valores qi, qj, pi, pj ingresados son en la cara del brazo rigido
    # por lo que se debria de crar un nuevo omdelo de carga PartialLoad pero en esta ocacion solo modificaremos el vector de cargas
    if qla == False:
        qa = qi
        pa = pi
    if qlb == False:
        qb = qj
        pb = pj
    ####################

    q = q_phi(L - la - lb, phi, qa, qb, pa, pb)   # Con deformaciones por corte (parte flexible)
    if qla:
        qla = q_phi(la, 0, qi, qa, pi, pa)            # Sin deformaciones por corte (parte rígida inicial)
    else:
        qla = np.zeros(6)
    if qlb:
        qlb = q_phi(lb, 0, qb, qj, pb, pj)            # Sin deformaciones por corte (parte rígida final)
    else:
        qlb = np.zeros(6)

    q_contacto = np.concatenate([qla[3:], qlb[:3]])   # Empotramiento en los contactos
    q_extremos = np.concatenate([qla[:3], qlb[3:]])   # Empotramiento en los extremos

    H = length_offset_transformation_matrix(la, lb)
    Q = np.dot(H.T, q + q_contacto) + q_extremos
    return Q


@lru_cache(maxsize=CACHE_COPIES)
def local_load_vector(
    L: float,  # Longitud del elemento
    phi: float,  # Aporte de cortante
    q_i: float,  # intensidad de carga transversal en el nodo i
    q_j: float,  # intensidad de carga transversal en el nodo j
    p_i: float,  # intensidad de carga axial en el nodo i
    p_j: float,  # intensidad de carga axial en el nodo j
) -> np.ndarray:
    """
    Calcula el vector de carga local para un elemento de longitud L
    con un coeficiente de Timoshenko phi y cargas distribuidas q_i y q_j.

    Args:
        L (float): Longitud del elemento.
        phi (float): Aporte de cortante.
        q_i (float): Carga distribuida en el nodo inicial.
        q_j (float): Carga distribuida en el nodo final.
        p_i (float): Carga axial en el nodo inicial.
        p_j (float): Carga axial en el nodo final.

    Returns:
        np.ndarray: Vector de carga local.
    """
    A = (q_j - q_i) / L
    B = q_i

    M = np.array([
        [L**2/2, L],
        [L**3 * (2 - phi) / 12, L**2 / 2]
    ])

    N = np.array([
        A * L**4 / 24 + B * L**3 / 6,
        A * L**5 * (0.6 - phi) / 72 + B * L**4 * (1 - phi) / 24
    ])

    C = np.linalg.solve(M, N)  # Más eficiente que inv(M) @ N

    Q = np.array([
        (2 * p_i + p_j) * L / 6,
        C[0],
        -C[1],
        (p_i + 2 * p_j) * L / 6,
        -(-A * L**2 / 2 - B * L + C[0]),
        (-A * L**3 / 6 - B * L**2 / 2 + C[0] * L + C[1]),
    ])

    return Q


# otra alternativa sin deformaciones de cortante
def load_vector(L, q_i, q_j):
    """
    Calcula el vector de carga local para una viga con carga distribuida q_i y q_j en los nodos inicial y final respectivamente.

    Parameters
    ----------
    L : float
        Longitud de la viga.
    q_i : float
        Carga distribuida en el nodo inicial.
    q_j : float
        Carga distribuida en el nodo final.

    Returns
    -------
    np.ndarray
        Vector de carga local.
    """
    import numpy as np
    f = np.array([
        0,
        L/20 * (7*q_i + 3*q_j),
        L**2 * (q_i/20 + q_j/30),
        0,
        L/20 * (3*q_i + 7*q_j),
        -L**2 * (q_i/30 + q_j/20),
    ])
    return f




from typing import Dict, Optional, TYPE_CHECKING
import numpy as np
import warnings

from milcapy.loads import PointLoad, DistributedLoad, PrescribedDOF, CSTLoad
from milcapy.utils import CoordinateSystemType, StateType, DirectionType, LoadType

if TYPE_CHECKING:
    from milcapy.model.model import SystemMilcaModel


def loads_to_global_system(load: PointLoad, angle: float) -> PointLoad:
    """
    Transforma una carga puntual del sistema local al sistema global.

    Args:
        load: Carga puntual a transformar.
        angle: Ángulo de rotación en radianes.

    Returns:
        PointLoad: Carga transformada al sistema global.
    """
    cos_a, sin_a = np.cos(angle), np.sin(angle)
    return PointLoad(
        fx=load.fx * cos_a - load.fy * sin_a,
        fy=load.fx * sin_a + load.fy * cos_a,
        mz=load.mz
    )


def distributed_load_to_local_system(
    system: "SystemMilcaModel",
    load_start: float,
    load_end: float,
    csys: CoordinateSystemType,
    direction: DirectionType,
    load_type: LoadType,
    member_id: int
) -> DistributedLoad:
    """
    Transforma una carga distribuida del sistema global al sistema local del miembro.

    Args:
        system: Modelo estructural que contiene el miembro.
        load_start: Valor inicial de la carga distribuida.
        load_end: Valor final de la carga distribuida.
        csys: Sistema de coordenadas de la carga.
        direction: Dirección de aplicación de la carga.
        load_type: Tipo de carga (fuerza o momento).
        member_id: Identificador del miembro.

    Returns:
        DistributedLoad: Carga transformada al sistema local del miembro.

    Raises:
        ValueError: Si la dirección especificada no es válida.
    """
    if csys == CoordinateSystemType.LOCAL:
        if load_type == LoadType.FORCE:
            if direction == DirectionType.LOCAL_1:
                return DistributedLoad(
                    q_i=0, q_j=0,
                    p_i=load_start, p_j=load_end,
                    m_i=0, m_j=0
                )
            elif direction == DirectionType.LOCAL_2:
                return DistributedLoad(
                    q_i=load_start, q_j=load_end,
                    p_i=0, p_j=0,
                    m_i=0, m_j=0
                )
            else:
                raise ValueError(f"Dirección de carga no válida: {direction}")
        elif load_type == LoadType.MOMENT:
            if direction == DirectionType.LOCAL_3:
                return DistributedLoad(
                    q_i=0, q_j=0,
                    p_i=0, p_j=0,
                    m_i=load_start, m_j=load_end
                )
            else:
                raise ValueError(f"Dirección de carga no válida para momento: {direction}")

    elif csys == CoordinateSystemType.GLOBAL:
        angle = system.members.get(member_id).angle_x()
        cos_a, sin_a = np.cos(angle), np.sin(angle)
        li, lj = load_start, load_end

        if load_type == LoadType.FORCE:
            if direction == DirectionType.X:
                return DistributedLoad(
                    q_i=-li * sin_a, q_j=-lj * sin_a,
                    p_i=li * cos_a, p_j=lj * cos_a,
                    m_i=0, m_j=0
                )
            elif direction == DirectionType.Y:
                return DistributedLoad(
                    q_i=li * cos_a, q_j=lj * cos_a,
                    p_i=li * sin_a, p_j=lj * sin_a,
                    m_i=0, m_j=0
                )
            elif direction == DirectionType.GRAVITY:
                return DistributedLoad(
                    q_i=-li * cos_a, q_j=-lj * cos_a,
                    p_i=-li * sin_a, p_j=-lj * sin_a,
                    m_i=0, m_j=0
                )
            elif direction == DirectionType.X_PROJ:
                return DistributedLoad(
                    q_i=-li * sin_a * sin_a, q_j=-lj * sin_a * sin_a,
                    p_i=li * cos_a * sin_a, p_j=lj * cos_a * sin_a,
                    m_i=0, m_j=0
                )
            elif direction == DirectionType.Y_PROJ:
                return DistributedLoad(
                    q_i=li * cos_a * cos_a, q_j=lj * cos_a * cos_a,
                    p_i=li * sin_a * cos_a, p_j=lj * sin_a * cos_a,
                    m_i=0, m_j=0
                )
            elif direction == DirectionType.GRAVITY_PROJ:
                return DistributedLoad(
                    q_i=-li * cos_a * cos_a, q_j=-lj * cos_a * cos_a,
                    p_i=-li * sin_a * cos_a, p_j=-lj * sin_a * cos_a,
                    m_i=0, m_j=0
                )
            else:
                raise ValueError(f"Dirección de carga no válida: {direction}")

        elif load_type == LoadType.MOMENT and direction == DirectionType.MOMENT:
            return DistributedLoad(
                q_i=0, q_j=0,
                p_i=0, p_j=0,
                m_i=li, m_j=lj
            )
        else:
            raise ValueError(f"Combinación no válida de tipo de carga y dirección: {load_type}, {direction}")


class LoadPattern:
    """
    Representa un patrón de carga en un modelo estructural.

    Esta clase gestiona la aplicación y transformación de cargas puntuales y distribuidas
    en un sistema estructural, permitiendo su manipulación en diferentes sistemas de coordenadas.
    """

    def __init__(
        self,
        name: str,
        self_weight_multiplier: float|None = None,
        state: StateType = StateType.ACTIVE,
        system: Optional["SystemMilcaModel"] = None,
    ) -> None:
        """
        Inicializa un nuevo patrón de carga con los parámetros especificados.

        Args:
            name: Nombre identificativo del patrón de carga.
            self_weight_multiplier: Factor multiplicador para el peso propio.
            state: Estado inicial del patrón de carga.
            system: Sistema estructural al que pertenece el patrón de carga.

        Raises:
            ValueError: Si el nombre está vacío o el multiplicador es negativo.
        """
        if not name.strip():
            raise ValueError("El nombre del patrón de carga no puede estar vacío")
        if self_weight_multiplier < 0:
            raise ValueError("El multiplicador de peso propio no puede ser negativo")

        self._system = system

        self.name = name
        self.self_weight_multiplier = self_weight_multiplier
        self.state = state
        self.analyzed = False

        self.point_loads: Dict[int, PointLoad] = {}
        self.distributed_loads: Dict[int, DistributedLoad] = {}
        self.prescribed_dof: Dict[int, PrescribedDOF] = {}
        self.cst_loads: Dict[int, CSTLoad] = {}

    def add_point_load(
        self,
        node_id: int,
        forces: PointLoad,
        csys: CoordinateSystemType = CoordinateSystemType.GLOBAL,
        angle_rot: Optional[float] = None,
        replace: bool = False,
    ) -> None:
        """
        Agrega o actualiza una carga puntual en un nodo específico.

        Args:
            node_id: Identificador del nodo objetivo.
            forces: Carga puntual a aplicar.
            csys: Sistema de coordenadas de la carga.
            angle_rot: Ángulo de rotación en radianes (solo para sistema LOCAL).
            replace: Si True, reemplaza cualquier carga existente en el nodo.

        Raises:
            TypeError: Si forces no es del tipo PointLoad.
            ValueError: Si el sistema de coordenadas es inválido o el ángulo es incorrecto.
        """
        if not isinstance(forces, PointLoad):
            raise TypeError("forces debe ser una instancia de PointLoad")

        if csys not in (CoordinateSystemType):
            raise ValueError("Sistema de coordenadas debe ser GLOBAL o LOCAL")

        transformed_forces = forces
        if csys == CoordinateSystemType.GLOBAL:
            if angle_rot is not None:
                warnings.warn("El ángulo de rotación se ignora en sistema GLOBAL")
        elif csys == CoordinateSystemType.LOCAL:
            if angle_rot is None:
                raise ValueError("Se debe indicar el ángulo de rotación en sistema LOCAL")
            transformed_forces = loads_to_global_system(forces, angle_rot)

        if replace:
            self.point_loads[node_id] = transformed_forces
        else:
            existing_load = self.point_loads.get(node_id, PointLoad())
            self.point_loads[node_id] = existing_load + transformed_forces

    def add_distributed_load(
        self,
        member_id: int,
        load_start: float,
        load_end: float,
        csys: CoordinateSystemType = CoordinateSystemType.LOCAL,
        direction: DirectionType = DirectionType.LOCAL_2,
        load_type: LoadType = LoadType.FORCE,
        replace: bool = False,
    ) -> None:
        """
        Agrega o actualiza una carga distribuida en un miembro específico.

        Args:
            member_id: Identificador del miembro objetivo.
            load_start: Valor inicial de la carga distribuida.
            load_end: Valor final de la carga distribuida.
            csys: Sistema de coordenadas de la carga.
            direction: Dirección de aplicación de la carga.
            load_type: Tipo de carga (fuerza o momento).
            replace: Si True, reemplaza cualquier carga existente en el miembro.

        Raises:
            ValueError: Si el sistema de coordenadas es inválido.
        """
        if self._system is None:
            raise ValueError("Se requiere un sistema válido para transformar cargas distribuidas")

        transformed_load = distributed_load_to_local_system(
            system=self._system,
            load_start=load_start,
            load_end=load_end,
            csys=csys,
            direction=direction,
            load_type=load_type,
            member_id=member_id
        )

        if replace:
            self.distributed_loads[member_id] = transformed_load
        else:
            existing_load = self.distributed_loads.get(member_id, DistributedLoad())
            self.distributed_loads[member_id] = existing_load + transformed_load

    def add_prescribed_dof(
        self,
        node_id: int,
        prescribed_dof: PrescribedDOF,
        replace: bool = False,
    ) -> None:
        """
        Agrega o actualiza un desplazamiento prescindido en un nodo específico.

        Args:
            node_id: Identificador del nodo objetivo.
            prescribed_dof: Desplazamiento prescindido a aplicar.
            replace: Si True, reemplaza cualquier desplazamiento prescindido existente en el nodo.

        Raises:
            TypeError: Si prescribed_dof no es del tipo PrescribedDOF.
        """
        if not isinstance(prescribed_dof, PrescribedDOF):
            raise TypeError("prescribed_dof debe ser una instancia de PrescribedDOF")

        if replace:
            self.prescribed_dof[node_id] = prescribed_dof
        else:
            existing_dof = self.prescribed_dof.get(node_id, PrescribedDOF())
            self.prescribed_dof[node_id] = existing_dof + prescribed_dof

    def add_cst_uniform_temperature_load(self, cst_id: int, dt: float) -> None:
        """
        agrega un incremento de temperatura al elemento CST
        """
        if self._system is None:
            raise ValueError("Se requiere un sistema válido para transformar cargas distribuidas")

        cst = self._system.csts.get(cst_id)
        if cst is None:
            raise ValueError(f"CST {cst_id} no encontrado en el sistema")

        if self.cst_loads.get(cst_id) is None:
            self.cst_loads[cst_id] = self.__create_cst_load(cst_id)

        self.cst_loads[cst_id].add_uniform_temperature_load(dt)

    def add_cst_uniform_distributed_load(self, cst_id: int, qx: float, qy: float) -> None:
        """
        agrega una carga distribuida uniforme superficial en su plano del elemento CST
        """
        if self._system is None:
            raise ValueError("Se requiere un sistema válido para transformar cargas distribuidas")

        cst = self._system.csts.get(cst_id)
        if cst is None:
            raise ValueError(f"CST {cst_id} no encontrado en el sistema")

        if self.cst_loads.get(cst_id) is None:
            self.cst_loads[cst_id] = self.__create_cst_load(cst_id)

        self.cst_loads[cst_id].add_uniform_distributed_load(qx, qy)

    def add_cst_uniform_edge_load(self, cst_id: int, q: float, edge: int) -> None:
        """
        agrega una carga distribuida uniforme en el borde del elemento CST
        """
        if self._system is None:
            raise ValueError("Se requiere un sistema válido para transformar cargas distribuidas")

        cst = self._system.csts.get(cst_id)
        if cst is None:
            raise ValueError(f"CST {cst_id} no encontrado en el sistema")

        if self.cst_loads.get(cst_id) is None:
            self.cst_loads[cst_id] = self.__create_cst_load(cst_id)

        self.cst_loads[cst_id].add_uniform_edge_load(q, edge)

    def add_cst_linear_edge_load(self, cst_id: int, qi: float, qj: float, edge: int) -> None:
        """
        agrega una carga distribuida lineal en el borde del elemento CST
        """
        if self._system is None:
            raise ValueError("Se requiere un sistema válido para transformar cargas distribuidas")

        cst = self._system.csts.get(cst_id)
        if cst is None:
            raise ValueError(f"CST {cst_id} no encontrado en el sistema")

        if self.cst_loads.get(cst_id) is None:
            self.cst_loads[cst_id] = self.__create_cst_load(cst_id)

        self.cst_loads[cst_id].add_linear_edge_load(qi, qj, edge)

    def __create_cst_load(self, cst_id: int) -> CSTLoad:
        """
        Crea un objeto CSTLoad para el elemento CST especificado.
        """
        if self._system is None:
            raise ValueError("Se requiere un sistema válido para transformar cargas distribuidas")

        cst = self._system.csts.get(cst_id)
        if cst is None:
            raise ValueError(f"CST {cst_id} no encontrado en el sistema")

        cd1 = cst.node1.vertex.coordinates
        cd2 = cst.node2.vertex.coordinates
        cd3 = cst.node3.vertex.coordinates
        E = cst.section.E()
        v = cst.section.v()
        t = cst.section.t
        beta = cst.section.beta()
        return CSTLoad(np.array([cd1, cd2, cd3]), E, v, t, beta)

    def assign_loads_to_nodes(self) -> None:
        """
        Asigna las cargas puntuales a los nodos correspondientes del sistema.
        """
        for node_id, load in self.point_loads.items():
            node = self._system.nodes.get(node_id)
            if node:
                node.set_load(load)
            else:
                warnings.warn(f"Nodo {node_id} no encontrado en el sistema")

    def assign_loads_to_members(self) -> None:
        """
        Asigna las cargas distribuidas a los miembros correspondientes del sistema.
        """
        for member_id, load in self.distributed_loads.items():
            members = self._system.members | self._system.trusses
            member = members.get(member_id)
            if member:
                member.set_distributed_load(load)
            else:
                warnings.warn(f"Miembro {member_id} no encontrado en el sistema")

        for cst_id, load in self.cst_loads.items():
            cst = self._system.csts.get(cst_id)
            if cst:
                cst.set_load(load)
            else:
                warnings.warn(f"CST {cst_id} no encontrado en el sistema")

    def assign_prescribed_dofs_to_nodes(self) -> None:
        """
        Asigna los desplazamientos prescindidos a los nodos correspondientes del sistema.
        """
        for node_id, dof in self.prescribed_dof.items():
            node = self._system.nodes.get(node_id)
            if node:
                node.set_prescribed_dof(dof)
            else:
                warnings.warn(f"Nodo {node_id} no encontrado en el sistema")


import numpy as np
from typing import TYPE_CHECKING
from matplotlib.patches import FancyArrowPatch
from milcapy.utils import rotate_xy, vertex_range


if TYPE_CHECKING:
    from matplotlib.pyplot import Axes

def _correction_angle(angle: float) -> float:
    """
    Corrige el angulo para que se dibuje correctamente
    """
    if 0 <= angle <= 90 or 270 < angle < 360:
        return angle
    elif 90 < angle <= 270:
        return (angle - 180)


def graphic_one_arrow(
    x: float,
    y: float,
    load: float,
    length_arrow: float,
    angle: float,
    ax: "Axes",
    color: str = "blue",
    label: bool = True,
    color_label: str = "black",
    label_font_size: int = 8,
    lw: float = 1,
    arrowstyle: str = "->"
) -> None:
    """
    Dibuja una flecha en un punto.
    """
    a = np.array([x, y])
    b = np.array([x + length_arrow * np.cos(angle), y + length_arrow * np.sin(angle)])

    # coordenadas al 15% de la punta de la flecha
    coord15p = np.array([x + 0.85 * length_arrow * np.cos(angle), y + 0.85 * length_arrow * np.sin(angle)])
    arrow = FancyArrowPatch(
        b, a,
        transform=ax.transData,
        color=color,
        linewidth=lw,
        arrowstyle=arrowstyle,
        mutation_scale=10,
        zorder=80
    )
    ax.add_patch(arrow)


    if label:
        text =ax.text(
            coord15p[0], coord15p[1],
            f"{redondear_si_mas_de_3_decimales(abs(load))}",
            fontsize=label_font_size,
            ha="right",
            va="bottom",
            color=color_label,
            rotation= _correction_angle(angle*180/np.pi),
            zorder=80
        )
    return arrow, text

def graphic_one_arrow_dof(
    x: float,
    y: float,
    load: float,
    length_arrow: float,
    angle: float,
    ax: "Axes",
    color: str = "blue",
    label: bool = True,
    color_label: str = "black",
    label_font_size: int = 8,
    lw: float = 1,
    arrowstyle: str = "->"
) -> None:
    """
    Dibuja una flecha en un punto.
    """
    a = np.array([x, y])
    b = np.array([x - length_arrow * np.cos(angle), y - length_arrow * np.sin(angle)])

    # coordenadas al 15% de la punta de la flecha
    coord15p = np.array([x - 0.85 * length_arrow * np.cos(angle), y - 0.85 * length_arrow * np.sin(angle)])
    arrow = FancyArrowPatch(
        b, a,
        transform=ax.transData,
        color=color,
        linewidth=lw,
        arrowstyle=arrowstyle,
        mutation_scale=10,
        zorder=80
    )
    ax.add_patch(arrow)


    if label:
        text =ax.text(
            coord15p[0], coord15p[1],
            f"{redondear_si_mas_de_3_decimales(abs(load))}",
            fontsize=label_font_size,
            ha="right",
            va="bottom",
            color=color_label,
            rotation= _correction_angle(angle*180/np.pi),
            zorder=80
        )
    return arrow, text



# import matplotlib.pyplot as plt
# fig, ax = plt.subplots()

# x = 0.65
# y = 0.5
# load = 400
# length_arrow = 0.5
# angle = np.pi/180 * 180
# graphic_one_arrow(x, y, load, length_arrow, angle, ax)
# plt.show()


# """
# si: X -> (-)=0, (+)=180
# si: Y -> (-)=90, (+)=270
# """


def graphic_n_arrow(
    x: float,
    y: float,
    load_i: float,
    load_j: float,
    angle: float,
    length: float,
    ax: "Axes",
    ratio_scale: float,
    nrof_arrows: int,
    color: str = "blue",
    angle_rotation: float = 0,
    label: bool = True,
    color_label: str = 'blue',
    label_font_size: int = 8,
    length_arrow = None
) -> None:
    """
    Dibuja una flecha en un punto.
    """
    # load_i, load_j = -load_i, -load_j
    # coordenadas de los extremos de la barra
    a = rotate_xy(np.array([x, y]), angle_rotation, x, y)
    b = rotate_xy(np.array([x + length, y]), angle_rotation, x, y)

    # coordenadas de los extremos carga
    if length_arrow:
        li = length_arrow*np.sign(load_i)
        lj = length_arrow*np.sign(load_j)
        c = rotate_xy(np.array([x + li * np.cos(angle), y + li * np.sin(angle)]), angle_rotation, x, y)
        d = rotate_xy(np.array([x + lj * np.cos(angle) + length, y + lj * np.sin(angle)]), angle_rotation, x, y)
    else:
        c = rotate_xy(np.array([x + load_i * ratio_scale * np.cos(angle), y + load_i * ratio_scale * np.sin(angle)]), angle_rotation, x, y)
        d = rotate_xy(np.array([x + load_j * ratio_scale * np.cos(angle) + length, y + load_j * ratio_scale * np.sin(angle)]), angle_rotation, x, y)

    cood_i = vertex_range(c, d, nrof_arrows)
    cood_j = vertex_range(a, b, nrof_arrows)

    # dibujar flechas
    arrows = []
    for start, end in zip(cood_i, cood_j):
        arrow = FancyArrowPatch(
            start, end,
            transform=ax.transData,
            color=color,
            linewidth=0.7,
            arrowstyle="->",
            mutation_scale=7,
            zorder=80
        )
        ax.add_patch(arrow)
        arrows.append(arrow)
    # linea que une las flechas
    line =ax.plot([c[0], d[0]], [c[1], d[1]], linewidth=0.7, color=color, zorder=80)

    # texto de la flecha
    texts = []
    if label:
        if load_i == load_j:
            coord_label = (c + d) / 2
            text = ax.text(
                coord_label[0], coord_label[1],
                f"{abs(load_i)}",
                fontsize=label_font_size,
                # ha="center",
                # va="center",
                color=color_label,
                zorder=80
            )
            texts.append(text)
        else:
            coord_label_i = c
            coord_label_j = d
            if load_i == 0:
                pass
            else:
                text = ax.text(
                    coord_label_i[0], coord_label_i[1],
                    f"{abs(load_i)}",
                    fontsize=label_font_size,
                    # ha="center",
                    # va="center",
                    color=color_label,
                    zorder=80
                )
                texts.append(text)
            if load_j == 0:
                pass
            else:
                text = ax.text(
                    coord_label_j[0], coord_label_j[1],
                    f"{abs(load_j)}",
                    fontsize=label_font_size,
                    # ha="center",
                    # va="center",
                    color=color_label,
                    zorder=80
                )
                texts.append(text)

        arrows.extend(line)  # Desempaqueta la lista de Line2D en arrows

        return arrows, texts


# x = 0
# y = 0
# load_i = 2
# load_j = 1
# length = 10
# angle = np.pi/2
# ratio_scale = 1.0
# nrof_arrows = 10
# color = 'blue'
# angle_rotation = 45
# label = True
# color_label = 'blue'
# label_font_size = 8

# import matplotlib.pyplot as plt
# fig, ax = plt.subplots()

# graphic_n_arrow(
#     x, y, load_i, load_j, angle, length, ax, ratio_scale, nrof_arrows, color, angle_rotation, label, color_label, label_font_size)
# plt.axis("equal")
# plt.show()

"""
si: X -> (-)=0, (+)=180
si: Y -> (-)=90, (+)=270
"""


def redondear_si_mas_de_3_decimales(num):
    """
    Redondea un número a 3 decimales solo si:
      - Es float, y
      - Tiene más de 3 decimales.
    Si es entero, lo devuelve como int.
    Elimina ceros innecesarios en los decimales.
    """
    # Si es equivalente a entero (int o float sin decimales)
    if float(num).is_integer():
        return int(num)

    # Si es float, contamos decimales reales
    num_str = f"{num:.16f}".rstrip("0").rstrip(".")
    if '.' in num_str:
        decimales = len(num_str.split('.')[1])
        if decimales > 3:
            num = round(num, 3)

    # Convertimos a str y quitamos ceros innecesarios
    return float(f"{num:.16f}".rstrip("0").rstrip("."))




def moment_fancy_arrow(
    ax: "Axes",
    x: float,
    y: float,
    moment: float,
    radio: float,
    color: str = 'blue',
    clockwise: bool = True,
    label: bool = True,
    color_label: str = 'blue',
    label_font_size: int = 8,
    lw: float = 1,
    arrowstyle: str = "->",
    ) :
    """
    Dibuja una flecha arco en un punto.
    """
    r = radio
    if moment < 0:
        curvature: float = -1.0 if clockwise else 1.0
    else:
        curvature: float = 1.0 if clockwise else -1.0

    arrow = FancyArrowPatch(
        (x - r, y - r), (x + r, y + r),
        connectionstyle=f"arc3,rad={curvature}",
        arrowstyle=arrowstyle,
        color=color,
        lw=lw,
        mutation_scale=10,
        zorder=80
    )
    ax.add_patch(arrow)

    if moment < 0:
        pos_label = (x - np.cos(np.pi/4) *(1.2)* r, y + np.cos(np.pi/4) * (1.8)*r)
    else:
        pos_label = (x + np.cos(np.pi/4) *(1.8)* r, y - np.cos(np.pi/4) * (2.1)*r)
    if label:
        text = ax.text(
            pos_label[0], pos_label[1],
            f"{redondear_si_mas_de_3_decimales(abs(moment))}",
            fontsize=label_font_size,
            ha="right",
            va="bottom",
            color=color_label,
            zorder=80
        )
    return arrow, text



# import matplotlib.pyplot as plt
# fig, ax = plt.subplots()

# x = 0.65
# y = 0.5
# moment = -400
# radio = 0.1
# arrow, text = moment_fancy_arrow(ax, x, y, moment, radio)
# plt.show()







def analyze_areas(points):
    """
    Analiza las áreas de un polígono.
    """
    if not points:
        return []

    def get_sign(y):
        return 1 if y >= 0 else -1

    areas = []
    current_area = [points[0]]
    current_sign = get_sign(points[0][1])

    for i in range(1, len(points)):
        current_point = points[i-1]
        next_point = points[i]
        next_sign = get_sign(next_point[1])

        if current_sign == next_sign:
            current_area.append(next_point)
        else:
            # Calcular la intersección
            x1, y1 = current_point
            x2, y2 = next_point
            delta_y = y2 - y1
            t = (-y1) / delta_y
            x_intercept = x1 + t * (x2 - x1)
            intersection = [x_intercept, 0.0]

            current_area.append(intersection)
            areas.append(current_area)

            # Iniciar nueva área
            current_area = [intersection, next_point]
            current_sign = next_sign

    # Añadir el área actual restante
    areas.append(current_area)

    return areas

# import numpy as np
# import matplotlib.pyplot as plt


# # Función para rotar puntos 45°
# def rotate_45(points):
#     theta = np.radians(45)
#     rotation_matrix = np.array([
#         [np.cos(theta), -np.sin(theta)],
#         [np.sin(theta), np.cos(theta)]
#     ])
#     rotated = []
#     for point in points:
#         rotated_point = rotation_matrix @ np.array(point)
#         rotated.append(rotated_point)
#     return np.array(rotated)


# # Generar datos de ejemplo
# x = np.linspace(0, 10, 100)
# y = np.sin(x) * 2  # Función con cambios de signo
# points = np.column_stack((x, y)).tolist()

# # Analizar áreas
# areas = analyze_areas(points)

# # Configurar el plot
# plt.figure(figsize=(12, 6))

# # Plot original
# plt.subplot(1, 2, 1)
# plt.title("Original")
# for i, area in enumerate(areas):
#     area = np.array(area)
#     color = 'blue' if area[0,1] >= 0 else 'red'
#     plt.plot(area[:,0], area[:,1], color=color, marker='o', markersize=3)
#     plt.fill_between(area[:,0], area[:,1], 0, color=color, alpha=0.2)
# plt.axhline(0, color='black', linestyle='--')
# plt.grid(True)

# # Plot rotado 45°
# plt.subplot(1, 2, 2)
# plt.title("Rotado 45°")
# for i, area in enumerate(areas):
#     rotated_area = rotate_45(area)
#     color = 'blue' if area[0][1] >= 0 else 'red'
#     plt.plot(rotated_area[:,0], rotated_area[:,1], color=color, marker='o', markersize=3)
#     plt.fill(rotated_area[:,0], rotated_area[:,1], color=color, alpha=0.2)
# plt.axhline(0, color='black', linestyle='--')
# plt.grid(True)

# plt.tight_layout()
# plt.show()


from typing import Tuple
from enum import Enum, auto

# Definición de restricciones
Restraints = Tuple[bool, bool, bool]


class ShearCoefficientMethodType(Enum):
    """Métodos para el cálculo del coeficiente de corte."""
    TIMOSHENKO = 'TIMOSHENKO'
    COWPER = 'COWPER'
    NONE = 'NONE'


class MemberType(Enum):
    """Tipos de miembros en la estructura."""
    FRAME = 'FRAME'
    TRUSS = 'TRUSS'
    BEAM = 'BEAM'
    BRACE = 'BRACE'
    MEMBRANE_2DOF = 'MEMBRANE_2DOF'
    MEMBRANE_3DOF = 'MEMBRANE_3DOF'

class BeamTheoriesType(Enum):
    """Tipos de teorías de vigas."""
    TIMOSHENKO = 'TIMOSHENKO'
    EULER_BERNOULLI = 'EULER_BERNOULLI'

class CoordinateSystemType(Enum):
    """Tipos de sistemas de coordenadas."""
    GLOBAL = 'GLOBAL'
    LOCAL = 'LOCAL'

class DirectionType(Enum):
    """Direcciones en el sistema global y local."""
    X = 'X'
    Y = 'Y'
    X_PROJ = 'X_PROJ'
    Y_PROJ = 'Y_PROJ'
    GRAVITY = 'GRAVITY'
    GRAVITY_PROJ = 'GRAVITY_PROJ'
    MOMENT = 'MOMENT'
    LOCAL_1 = 'LOCAL_1' # AXIAL
    LOCAL_2 = 'LOCAL_2' # SHEAR
    LOCAL_3 = 'LOCAL_3' # MOMENT

class StateType(Enum):
    """Estado de los elementos o componentes."""
    ACTIVE = 'ACTIVE'
    INACTIVE = 'INACTIVE'


class LoadType(Enum):
    """Tipos de cargas (fuerzas o momentos)."""
    FORCE = 'FORCE'
    MOMENT = 'MOMENT'


class InternalForceType(Enum):
    """Enumeración de tipos de Fuerzas Internas."""
    AXIAL_FORCE = "AXIAL_FORCE"
    SHEAR_FORCE = "SHEAR_FORCE"
    BENDING_MOMENT = "BENDING_MOMENT"
    SLOPE = "SLOPE"
    DEFLECTION = "DEFLECTION"
    DEFORMED = "DEFORMED"
    RIGID_DEFORMED = "RIGID_DEFORMED"



class FieldTypeMembrane(Enum):
    SX = "SX"
    SY = "SY"
    SXY = "SXY"
    EX = "EX"
    EY = "EY"
    EXY = "EXY"
    UX = "UX"
    UY = "UY"
    UMAG = "UMAG"



def to_enum(key: str, enum: Enum) -> Enum:
    """Convierte un string a un miembro de un Enum."""
    assert isinstance(key, str), 'La clave debe ser un string.'
    assert issubclass(enum, Enum), 'El segundo argumento debe ser una clase Enum.'
    try:
        return enum(key)
    except ValueError:
        raise ValueError(f'La clave "{key}" no se encuentra en el Enum "{enum.__name__}".')


class ConstitutiveModel(Enum):
    PLANE_STRESS = 'PLANE_STRESS'     # Esfuerzo plano
    PLANE_STRAIN = 'PLANE_STRAIN'     # Deformación plana



class IntegrationType(Enum):
    """Tipos de integración."""
    COMPLETE = 'COMPLETE'
    REDUCED = 'REDUCED'


class MembraneQuadElementType(Enum):
    """Tipos de elementos de membrana."""
    MQ4 = auto()
    MQ6 = auto()
    MQ6I = auto()
    MQ8Reduced = auto()
    MQ8Complete = auto()



import numpy as np
from typing import Union
from milcapy.core.node import Node
from milcapy.section.section import ShellSection
from milcapy.utils.types import ConstitutiveModel
from milcapy.elements.quad4 import MembraneQuad4
from milcapy.elements.MQ6   import MembraneQuad6
from milcapy.elements.MQ6I  import MembraneQuad6I
from milcapy.elements.quad8 import MembraneQuad8
from milcapy.utils.types    import MembraneQuadElementType, IntegrationType


class MembraneQuad6IMod(MembraneQuad6):
    def __init__(
        self,
        id: int,
        node1: Node,
        node2: Node,
        node3: Node,
        node4: Node,
        section: ShellSection,
        state: Union[ConstitutiveModel, str] = ConstitutiveModel.PLANE_STRESS,
        ele_type: Union[MembraneQuadElementType, str] = MembraneQuadElementType.MQ6,
    ):
        super().__init__(id, node1, node2, node3, node4, section, state=state)

    def K_global(self) -> np.ndarray:
        """
        Matriz de rigidez global.
        """

        if self.ele_type == MembraneQuadElementType.MQ4:
            MQ4 = MembraneQuad4(self.id, self.node1, self.node2, self.node3, self.node4, self.section, state=self.state) # tiene rigidez de desplazamientos
            Kdisp = MQ4.global_stiffness_matrix()
        elif self.ele_type == MembraneQuadElementType.MQ6:
            return super().global_stiffness_matrix()
        elif self.ele_type == MembraneQuadElementType.MQ6I:
            MQ6I = MembraneQuad6I(self.id, self.node1, self.node2, self.node3, self.node4, self.section, state=self.state) # tiene rigidez de desplazamientos
            Kdisp = MQ6I.global_stiffness_matrix()
        elif self.ele_type == MembraneQuadElementType.MQ8Reduced:
            MQ8 = MembraneQuad8(self.id, self.node1, self.node2, self.node3, self.node4, self.section, state=self.state, integration_type=IntegrationType.REDUCED) # tiene rigidez de desplazamientos
            Kdisp = MQ8.global_stiffness_matrix()
        elif self.ele_type == MembraneQuadElementType.MQ8Complete:
            MQ8 = MembraneQuad8(self.id, self.node1, self.node2, self.node3, self.node4, self.section, state=self.state, integration_type=IntegrationType.COMPLETE) # tiene rigidez de desplazamientos
            Kdisp = MQ8.global_stiffness_matrix()

        Krot = super().global_stiffness_matrix()
        # Krot = self.Ki()
        # Kdisp = MQ6I.Ki()
        # T = self.get_transformation_matrix()
        dofDisp = [0, 1,      3, 4,     6, 7,     9, 10]
        # INSERTAMOS LOS K DISP EN LOS DOFDISP DE KROT
        Krot[np.ix_(dofDisp,dofDisp)] = Kdisp
        K = Krot
        # K_GLOBAL = T @ K @ T.T
        return K

    def force_vector(self) -> np.ndarray:
        """
        Vector de fuerzas nodales.
        """
        pass
